"""CLI of application module."""

import json
import sys
import time
import zipfile
from pathlib import Path
from typing import Annotated

import typer

from aignostics.bucket import Service as BucketService
from aignostics.platform import NotFoundException
from aignostics.utils import console, get_logger, get_user_data_directory, sanitize_path

from ._service import DownloadProgress, DownloadProgressState, Service
from ._utils import (
    application_run_status_to_str,
    print_runs_non_verbose,
    print_runs_verbose,
    read_metadata_csv_to_dict,
    retrieve_and_print_run_details,
    write_metadata_dict_to_csv,
)

MESSAGE_NOT_YET_IMPLEMENTED = "NOT YET IMPLEMENTED"

logger = get_logger(__name__)

cli = typer.Typer(name="application", help="List and inspect applications on Aignostics Platform.")

run_app = typer.Typer()
cli.add_typer(run_app, name="run", help="List, submit and manage application runs")

result_app = typer.Typer()
run_app.add_typer(result_app, name="result", help="Download or delete run results.")


@cli.command("list")
def application_list(
    verbose: Annotated[bool, typer.Option(help="Show application details")] = False,
) -> None:
    """List available applications."""
    try:
        applications = Service().applications()
    except Exception as e:
        logger.exception("Failed to list applications")
        console.print(f"[error]Error:[/error] Failed to list applications: {e}")
        sys.exit(1)

    app_count = 0

    if verbose:
        console.print("[bold]Available Applications:[/bold]")
        console.print("=" * 80)

        for app in applications:
            app_count += 1
            console.print(f"[bold]Application ID:[/bold] {app.application_id}")
            console.print(f"[bold]Name:[/bold] {app.name}")
            console.print(f"[bold]Regulatory Classes:[/bold] {', '.join(app.regulatory_classes)}")

            try:
                versions = Service().application_versions(app)
            except Exception as e:
                logger.exception("Failed to list versions for application '%s'", app.application_id)
                console.print(
                    f"[error]Error:[/error] Failed to list versions for application '{app.application_id}': {e}"
                )
                continue
            if versions:
                console.print("[bold]Available Versions:[/bold]")
                for version in versions:
                    console.print(f"  - {version.version} ({version.application_version_id})")
                    console.print(f"    Changelog: {version.changelog}")

                    num_inputs = len(version.input_artifacts)
                    num_outputs = len(version.output_artifacts)
                    console.print(f"    Artifacts: {num_inputs} input(s), {num_outputs} output(s)")

            console.print("[bold]Description:[/bold]")
            for line in app.description.strip().split("\n"):
                console.print(f"  {line}")

            console.print("-" * 80)
    else:
        console.print("[bold]Available Aignostics Applications:[/bold]")
        for app in applications:
            app_count += 1
            latest_version = Service().application_version_latest(app)
            console.print(
                f"- [bold]{app.application_id}[/bold] - latest application version id: "
                f"`{latest_version.application_version_id if latest_version else 'None'}`"
            )

    if app_count == 0:
        logger.info("No applications available.")
        console.print("No applications available.")


@cli.command("dump-schemata")
def application_dump_schemata(  # noqa: C901
    id: Annotated[  # noqa: A002
        str,
        typer.Argument(
            help="Id of the application or application_version to dump the output schema for. "
            "If application id is given the latest version of the application will be used."
        ),
    ],
    destination: Annotated[
        Path,
        typer.Option(
            help="Path pointing to directory where the input and output schemata will be dumped.",
            exists=False,
            file_okay=False,
            dir_okay=True,
            writable=True,
            readable=True,
            resolve_path=True,
        ),
    ] = Path().cwd(),  # noqa: B008,
    zip: Annotated[  # noqa: A002
        bool,
        typer.Option(
            help="If set, the schema files will be zipped into a single file, with the schema files deleted.",
        ),
    ] = False,
) -> None:
    """Output the input schema of the application in JSON format."""
    try:
        application_version = Service().application_version(id, True)
        application = Service().application(application_version.application_id)
    except (NotFoundException, ValueError) as e:
        message = f"Failed to load application version with ID '{id}', check your input: : {e!s}."
        logger.warning(message)
        console.print(f"[warning]Warning:[/warning] {message}")
        sys.exit(2)
    except (Exception, RuntimeError) as e:
        message = f"Failed to load application version with ID '{id}': {e!s}."
        logger.exception(message)
        console.print(f"[warning]Error:[/warning] {message}")
        sys.exit(1)
    try:
        destination.mkdir(parents=True, exist_ok=True)
    except OSError as e:
        console.print(f"[error]Error:[/error] Failed to create directory '{destination}': {e}")
        sys.exit(1)

    created_files: list[Path] = []

    for input_artifact in application_version.input_artifacts:
        if input_artifact.metadata_schema:
            file_path: Path = sanitize_path(
                Path(destination / f"{application_version.application_version_id}_input_{input_artifact.name}.json")
            )  # type: ignore
            file_path.write_text(data=json.dumps(input_artifact.metadata_schema, indent=2), encoding="utf-8")
            created_files.append(file_path)

    for output_artifact in application_version.output_artifacts:
        if output_artifact.metadata_schema:
            file_path = sanitize_path(
                Path(destination / f"{application_version.application_version_id}_output_{output_artifact.name}.json")
            )  # type: ignore
            file_path.write_text(data=json.dumps(output_artifact.metadata_schema, indent=2), encoding="utf-8")
            created_files.append(file_path)

    md_file_path: Path = sanitize_path(Path(destination / f"{application_version.application_version_id}_schemata.md"))  # type: ignore
    with md_file_path.open("w", encoding="utf-8") as md_file:
        md_file.write(f"# Schemata for Aignostics Application {application.name}\n")
        md_file.write(f"* ID: {application.application_id}\n")
        md_file.write(f"* Version ID: {application_version.application_version_id}\n")
        md_file.write(f"\n## Description: \n{application.description}\n\n")
        md_file.write("\n## Input Artifacts\n")
        for input_artifact in application_version.input_artifacts:
            md_file.write(
                f"- {input_artifact.name}: "
                f"{application_version.application_version_id}_input_{input_artifact.name}.json\n"
            )
        md_file.write("\n## Output Artifacts\n")
        for output_artifact in application_version.output_artifacts:
            md_file.write(
                f"- {output_artifact.name}: "
                f"{application_version.application_version_id}_output_{output_artifact.name}.json\n"
            )
    created_files.append(md_file_path)

    if zip:
        zip_filename = sanitize_path(Path(destination / f"{application_version.application_version_id}_schemata.zip"))
        with zipfile.ZipFile(zip_filename, "w", zipfile.ZIP_DEFLATED) as zipf:
            for file_path in created_files:
                zipf.write(file_path, arcname=file_path.name)
        console.print(f"Zipped {len(created_files)} files to [bold]{zip_filename}[/bold]")
        for file_path in created_files:
            file_path.unlink()


@cli.command("describe")
def application_describe(
    application_id: Annotated[str, typer.Argument(help="Id of the application to describe")],
) -> None:
    """Describe application."""
    try:
        application = Service().application(application_id)
    except NotFoundException:
        logger.warning("Application with ID '%s' not found.", application_id)
        console.print(f"[warning]Warning:[/warning] Application with ID '{application_id}' not found.")
        sys.exit(2)
    except Exception as e:
        logger.exception("Failed to describe application with ID '%s'", application_id)
        console.print(f"[error]Error:[/error] Failed to describe application: {e}")
        sys.exit(1)

    console.print(f"[bold]Application Details for {application.application_id}[/bold]")
    console.print("=" * 80)
    console.print(f"[bold]Name:[/bold] {application.name}")
    console.print(f"[bold]Regulatory Classes:[/bold] {', '.join(application.regulatory_classes)}")

    console.print("[bold]Description:[/bold]")
    for line in application.description.strip().split("\n"):
        console.print(f"  {line}")

    versions = Service().application_versions(application)
    if versions:
        console.print()
        console.print("[bold]Available Versions:[/bold]")
        for version in versions:
            console.print(f"  [bold]Version ID:[/bold] {version.application_version_id}")
            console.print(f"  [bold]Version:[/bold] {version.version}")
            console.print(f"  [bold]Changelog:[/bold] {version.changelog}")

            console.print("  [bold]Input Artifacts:[/bold]")
            for artifact in version.input_artifacts:
                console.print(f"    - Name: {artifact.name}")
                console.print(f"      MIME Type: {artifact.mime_type}")
                console.print(f"      Schema: {artifact.metadata_schema}")

            console.print("  [bold]Output Artifacts:[/bold]")
            for artifact in version.output_artifacts:
                console.print(f"    - Name: {artifact.name}")
                console.print(f"      MIME Type: {artifact.mime_type}")
                console.print(f"      Scope: {artifact.scope}")
                console.print(f"      Schema: {artifact.metadata_schema}")

            console.print()


@run_app.command(name="execute")
def run_execute(  # noqa: PLR0913, PLR0917
    application_version_id: Annotated[
        str,
        typer.Argument(
            help="Id of application version to execute. "
            "If application id is given, the latest version of that application is used."
        ),
    ],
    metadata_csv_file: Annotated[
        Path,
        typer.Argument(
            help="Filename of the .csv file containing the metadata and references.",
            exists=False,
            file_okay=True,
            dir_okay=False,
            writable=True,
            readable=True,
            resolve_path=True,
        ),
    ],
    source_directory: Annotated[
        Path,
        typer.Argument(
            help="Source directory to scan for whole slide images",
            exists=True,
            file_okay=False,
            dir_okay=True,
            writable=True,
            readable=True,
            resolve_path=True,
        ),
    ],
    mapping: Annotated[
        list[str],
        typer.Argument(
            help="Mapping to use for amending metadata CSV file. "
            "Each mapping is of the form '<regexp>:<key>:<value>,<key>:<value>,...'."
            "The regular expression is matched against the reference attribute of the entry. "
            "The key/value pairs are applied to the entry if the pattern matches. "
            "You can use the mapping option multiple times to set values for multiple files. "
            'Example: ".*:staining_method:H&E,tissue:LIVER,disease:LIVER_CANCER"',
        ),
    ],
    create_subdirectory_for_run: Annotated[
        bool,
        typer.Option(
            help="Create a subdirectory for the results of the run in the destination directory",
        ),
    ] = True,
    create_subdirectory_per_item: Annotated[
        bool,
        typer.Option(
            help="Create a subdirectory per item in the destination directory",
        ),
    ] = True,
    upload_prefix: Annotated[
        str,
        typer.Option(
            help="Prefix for the upload destination. If not given will be set to current milliseconds.",
        ),
    ] = f"{time.time() * 1000}",
    wait_for_completion: Annotated[
        bool,
        typer.Option(
            help="Wait for run completion and download results incrementally",
        ),
    ] = True,
) -> None:
    """Prepare metadata, upload data to platform, and submit an application run, then incrementally download results.

    (1) Prepares metadata CSV file for the given application version
        by scanning the source directory for whole slide images
        and extracting metadata such as width, height, and mpp.
    (2) Amends the metadata CSV file using the given mappings
        to set all required attributes.
    (3) Uploads the files referenced in the metadata CSV file
        to the cloud bucket provisioned in the Aignostics platform.
    (4) Submits the run for the given application version
        with the metadata from the CSV file.
    (5) Downloads the results of the run to the destination directory,
        by default waiting for the run to complete
        and downloading results incrementally.
    """
    run_prepare(
        application_version_id=application_version_id,
        metadata_csv=metadata_csv_file,
        source_directory=source_directory,
        mapping=mapping,
    )
    run_upload(
        application_version_id=application_version_id,
        metadata_csv_file=metadata_csv_file,
        upload_prefix=upload_prefix,
    )
    application_run_id = run_submit(
        application_version_id=application_version_id,
        metadata_csv_file=metadata_csv_file,
    )
    result_download(
        run_id=application_run_id,
        destination_directory=metadata_csv_file.parent,
        create_subdirectory_for_run=create_subdirectory_for_run,
        create_subdirectory_per_item=create_subdirectory_per_item,
        wait_for_completion=wait_for_completion,
    )


@run_app.command(name="prepare")
def run_prepare(
    application_version_id: Annotated[
        str,
        typer.Argument(
            help="Id of the application version to generate the metadata for. "
            "If application id is given, the latest version of that application is used."
        ),
    ],
    metadata_csv: Annotated[
        Path,
        typer.Argument(
            help="Target filename for the generated metadata CSV file.",
            exists=False,
            file_okay=True,
            dir_okay=False,
            writable=True,
            readable=True,
            resolve_path=True,
        ),
    ],
    source_directory: Annotated[
        Path,
        typer.Argument(
            help="Source directory to scan for whole slide images",
            exists=True,
            file_okay=False,
            dir_okay=True,
            writable=True,
            readable=True,
            resolve_path=True,
        ),
    ],
    mapping: Annotated[
        list[str] | None,
        typer.Option(
            help="Mapping to use for amending metadata CSV file. "
            "Each mapping is of the form '<regexp>:<key>:<value>,<key>:<value>,...'. "
            "The regular expression is matched against the reference attribute of the entry. "
            "The key/value pairs are applied to the entry if the pattern matches. "
            "You can use the mapping option multiple times to set values for multiple files. "
        ),
    ] = None,
) -> None:
    """Prepare metadata CSV file required for submitting a run.

    (1) Scans source_directory for whole slide images.
    (2) Extracts metadata from whole slide images such as width, height, mpp.
    (3) Creates CSV file with columns as required by the given application version.
    (4) Optionally applies mappings to amend the metadata CSV file for columns
        that are not automatically filled by the metadata extraction process.

    Example:
        aignostics application run prepare "he-tme:v0.51.0" metadata.csv /path/to/source_directory
        --mapping "*.tiff:staining_method:H&E,tissue:LUNG,disease:LUNG_CANCER"
    """
    write_metadata_dict_to_csv(
        metadata_csv=metadata_csv,
        metadata_dict=Service().generate_metadata_from_source_directory(
            application_version_id=application_version_id,
            source_directory=source_directory,
            mappings=mapping or [],
        ),
    )
    console.print(f"Generated metadata file [bold]{metadata_csv}[/bold]")
    logger.info("Generated metadata file: '%s'", metadata_csv)


@run_app.command(name="upload")
def run_upload(
    application_version_id: Annotated[
        str,
        typer.Argument(
            help="Id of the application version to upload data for. "
            "If application id is given, the latest version of that application is used."
        ),
    ],
    metadata_csv_file: Annotated[
        Path,
        typer.Argument(
            help="Filename of the .csv file containing the metadata and references.",
            exists=True,
            file_okay=True,
            dir_okay=False,
            writable=True,
            readable=True,
            resolve_path=True,
        ),
    ],
    upload_prefix: Annotated[
        str,
        typer.Option(
            help="Prefix for the upload destination. If not given will be set to current milliseconds.",
        ),
    ] = f"{time.time() * 1000}",
) -> None:
    """Upload files referenced in the metadata CSV file to the Aignostics platform.

    1. Reads the metadata CSV file.
    2. Uploads the files referenced in the CSV file to the Aignostics platform.
    3. Incrementally updates the CSV file with upload progress and the signed URLs for the uploaded files.
    """
    from rich.progress import (  # noqa: PLC0415
        BarColumn,
        FileSizeColumn,
        Progress,
        TaskProgressColumn,
        TextColumn,
        TimeRemainingColumn,
        TotalFileSizeColumn,
        TransferSpeedColumn,
    )

    metadata_dict = read_metadata_csv_to_dict(metadata_csv_file=metadata_csv_file)
    if not metadata_dict:
        sys.exit(2)

    total_bytes = 0
    for i, entry in enumerate(metadata_dict):
        source = entry["reference"]
        source_file_path = Path(source)
        if not source_file_path.is_file():
            logger.warning("Source file '%s' (row %d) does not exist", source_file_path, i)
            console.print(f"[warning]Warning:[/warning] Source file '{source_file_path}' (row {i}) does not exist")
            sys.exit(2)

        total_bytes += source_file_path.stat().st_size

    with Progress(
        TextColumn(
            f"[progress.description]Uploading from {metadata_csv_file} to "
            f"{BucketService().get_bucket_protocol()}:/{BucketService().get_bucket_name()}/{upload_prefix}"
        ),
        BarColumn(),
        TaskProgressColumn(),
        TimeRemainingColumn(),
        FileSizeColumn(),
        TotalFileSizeColumn(),
        TransferSpeedColumn(),
        TextColumn("[progress.description]{task.description}"),
    ) as progress:
        task = progress.add_task(f"Uploading to {upload_prefix}/...", total=total_bytes)

        def update_progress(bytes_uploaded: int, source: Path, platform_bucket_url: str) -> None:
            progress.update(task, advance=bytes_uploaded, description=f"{source.name}")
            for entry in metadata_dict:
                if entry["reference"] == str(source):
                    entry["platform_bucket_url"] = platform_bucket_url
                    break
            write_metadata_dict_to_csv(
                metadata_csv=metadata_csv_file,
                metadata_dict=metadata_dict,
            )

        Service().application_run_upload(
            application_version_id=application_version_id,
            metadata=metadata_dict,
            upload_prefix=upload_prefix,
            upload_progress_callable=update_progress,
        )

    logger.info("Upload completed.")
    console.print("Upload completed.", style="info")


@run_app.command("submit")
def run_submit(
    application_version_id: Annotated[
        str,
        typer.Argument(
            help="Id of the application version to submit run for. "
            "If application id is given, the latest version of that application is used."
        ),
    ],
    metadata_csv_file: Annotated[
        Path,
        typer.Argument(
            help="Filename of the .csv file containing the metadata and references.",
            exists=False,
            file_okay=True,
            dir_okay=False,
            writable=False,
            readable=True,
            resolve_path=True,
        ),
    ],
) -> str:
    """Submit run by referencing the metadata CSV file.

    1. Requires the metadata CSV file to be generated and referenced files uploaded first

    Returns:
        The ID of the submitted application run.
    """
    try:
        metadata_dict = read_metadata_csv_to_dict(metadata_csv_file=metadata_csv_file)
        if not metadata_dict:
            console.print("Could mot read metadata file '%s'", metadata_csv_file)
            sys.exit(2)
        logger.debug(
            "Submitting run for application version '%s' with metadata: %s", application_version_id, metadata_dict
        )
        application_run = Service().application_run_submit_from_metadata(
            application_version_id=application_version_id,
            metadata=metadata_dict,
        )
        console.print(f"Submitted run with id '{application_run.application_run_id}' for '{application_version_id}'.")
        return application_run.application_run_id
    except ValueError as e:
        logger.warning("Bad input to create run for application version '%s': %s", application_version_id, e)
        console.print(
            f"[warning]Warning:[/warning] Bad input to create run for application version "
            f"'{application_version_id}': {e}"
        )
        sys.exit(2)
    except Exception as e:
        logger.exception("Failed to create run for application version '%s'", application_version_id)
        console.print(
            f"[error]Error:[/error] Failed to create run for application version '{application_version_id}': {e}"
        )
        sys.exit(1)


@run_app.command("list")
def run_list(
    verbose: Annotated[bool, typer.Option(help="Show application details")] = False,
    limit: Annotated[int | None, typer.Option(help="Maximum number of runs to display")] = None,
) -> None:
    """List runs."""
    try:
        runs = Service().application_runs(limit=limit)
        if len(runs) == 0:
            message = "You did not yet create a run."
            logger.warning(message)
            console.print(message, style="warning")
        else:
            print_runs_verbose(runs) if verbose else print_runs_non_verbose(runs)
            message = f"Listed '{len(runs)}' run(s)."
            console.print(message, style="info")
            logger.info(message, len(runs))
    except Exception as e:
        logger.exception("Failed to list runs")
        console.print(f"[error]Error:[/error] Failed to list runs: {e}")


@run_app.command("describe")
def run_describe(run_id: Annotated[str, typer.Argument(help="Id of the run to describe")]) -> None:
    """Describe run."""
    logger.debug("Describing run with ID '%s'", run_id)

    try:
        retrieve_and_print_run_details(Service().application_run(run_id))
        logger.info("Described run with ID '%s'", run_id)
    except NotFoundException:
        logger.warning("Run with ID '%s' not found.", run_id)
        console.print(f"[warning]Warning:[/warning] Run with ID '{run_id}' not found.")
        sys.exit(2)
    except Exception as e:
        logger.exception("Failed to retrieve and print run details for ID '%s'", run_id)
        console.print(f"[error]Error:[/error] Failed to retrieve run details for ID '{run_id}': {e}")
        sys.exit(1)


@run_app.command("cancel")
def run_cancel(
    run_id: Annotated[str, typer.Argument(..., help="Id of the run to cancel")],
) -> None:
    """Cancel run."""
    logger.debug("Canceling run with ID '%s'", run_id)

    try:
        Service().application_run_cancel(run_id)
        logger.info("Canceled run with ID '%s'.", run_id)
        console.print(f"Run with ID '{run_id}' has been canceled.")
    except NotFoundException:
        logger.warning("Run with ID '%s' not found.", run_id)
        console.print(f"[warning]Warning:[/warning] Run with ID '{run_id}' not found.")
        sys.exit(2)
    except Exception as e:
        logger.exception("Failed to cancel run with ID '%s'", run_id)
        console.print(f"[bold red]Error:[/bold red] Failed to cancel run with ID '{run_id}': {e}")
        sys.exit(1)


@result_app.command("download")
def result_download(  # noqa: PLR0913, PLR0917
    run_id: Annotated[str, typer.Argument(..., help="Id of the run to download results for")],
    destination_directory: Annotated[
        Path,
        typer.Argument(
            help="Destination directory to download results to",
            exists=False,
            file_okay=False,
            dir_okay=True,
            writable=True,
            readable=True,
            resolve_path=True,
        ),
    ] = get_user_data_directory("results"),  # noqa: B008
    create_subdirectory_for_run: Annotated[
        bool,
        typer.Option(
            help="Create a subdirectory for the results of the run in the destination directory",
        ),
    ] = True,
    create_subdirectory_per_item: Annotated[
        bool,
        typer.Option(
            help="Create a subdirectory per item in the destination directory",
        ),
    ] = True,
    wait_for_completion: Annotated[
        bool,
        typer.Option(
            help="Wait for run completion and download results incrementally",
        ),
    ] = True,
    qupath_project: Annotated[
        bool,
        typer.Option(
            help="Create a QuPath project referencing input slides and results. \n"
            "The QuPath project will be created in a subfolder of the destination directory. \n"
            "This option requires the QuPath extension for Launchpad: "
            'start the Launchpad with `uvx --with "aignostics[qupath]" aignostics ...` \n'
            "This options requires installation of the QuPath application: "
            'Run uvx --with "aignostics[qupath]" aignostics qupath install'
        ),
    ] = False,
) -> None:
    """Download results of a run."""
    logger.debug(
        "Downloading results for run with ID '%s' to '%s' with options: "
        "create_subdirectory_for_run=%s, create_subdirectory_per_item=%s, wait_for_completion=%s, qupath_project=%r",
        run_id,
        destination_directory,
        create_subdirectory_for_run,
        create_subdirectory_per_item,
        wait_for_completion,
        qupath_project,
    )
    from rich.console import Group  # noqa: PLC0415
    from rich.live import Live  # noqa: PLC0415
    from rich.panel import Panel  # noqa: PLC0415
    from rich.progress import (  # noqa: PLC0415
        BarColumn,
        FileSizeColumn,
        Progress,
        TaskID,
        TaskProgressColumn,
        TextColumn,
        TimeElapsedColumn,
        TimeRemainingColumn,
        TotalFileSizeColumn,
        TransferSpeedColumn,
    )

    try:
        download_tasks: dict[str, TaskID] = {}

        main_download_progress_ui = Progress(
            BarColumn(),
            TaskProgressColumn(),
            TextColumn("[progress.description]{task.description}"),
            TimeElapsedColumn(),
            TimeRemainingColumn(),
            TextColumn("{task.fields[extra_description]}"),
        )
        artifact_download_progress_ui = Progress(
            BarColumn(),
            TaskProgressColumn(),
            TextColumn("[progress.description]{task.description}"),
            TimeElapsedColumn(),
            TimeRemainingColumn(),
            FileSizeColumn(),
            TotalFileSizeColumn(),
            TransferSpeedColumn(),
            TextColumn("{task.fields[extra_description]}"),
        )
        panel = Panel(
            Group(main_download_progress_ui, artifact_download_progress_ui),
            title=f"Run {run_id}",
            subtitle="",
            highlight=True,
        )
        with Live(panel):
            main_task = main_download_progress_ui.add_task(description="", total=None, extra_description="")

            def update_progress(progress: DownloadProgress) -> None:
                """Update progress bar for file downloads."""
                if progress.run:
                    panel.title = f"Run {progress.run.application_run_id} of {progress.run.application_version_id}"
                    panel.subtitle = f"Triggered at {progress.run.triggered_at.strftime('%a, %x %X')}"
                    if progress.item_count:
                        panel.subtitle += f" with {progress.item_count} " + (
                            "item" if progress.item_count == 1 else "items"
                        )
                    panel.subtitle += f", status: {application_run_status_to_str(progress.run.status)}."
                main_download_progress_ui.update(
                    main_task,
                    description=(
                        progress.status
                        if progress.status is not DownloadProgressState.DOWNLOADING or not progress.total_artifact_index
                        else (
                            f"Downloading artifact {progress.total_artifact_index + 1} "
                            f"of {progress.total_artifact_count}"
                        )
                    ).ljust(50),
                )
                if progress.artifact_path:
                    task_key = str(progress.artifact_path.absolute())
                    if task_key not in download_tasks:
                        download_tasks[task_key] = artifact_download_progress_ui.add_task(
                            f"{progress.artifact_path.name}".ljust(50),
                            total=progress.artifact_size,
                            extra_description=f"Item {progress.item.reference if progress.item else 'unknown'}",
                        )

                    artifact_download_progress_ui.update(
                        download_tasks[task_key],
                        total=progress.artifact_size,
                        advance=progress.artifact_downloaded_chunk_size,
                    )

                if (
                    progress.item_count
                    and progress.item_index is not None
                    and progress.artifact_count
                    and progress.artifact_index is not None
                ):
                    main_download_progress_ui.update(
                        main_task,
                        completed=progress.item_index * progress.artifact_count + progress.artifact_index + 1,
                        total=float(progress.total_artifact_count) if progress.total_artifact_count else 0.0,
                    )

            destination_directory = Service().application_run_download(
                run_id=run_id,
                destination_directory=destination_directory,
                create_subdirectory_for_run=create_subdirectory_for_run,
                create_subdirectory_per_item=create_subdirectory_per_item,
                wait_for_completion=wait_for_completion,
                qupath_project=qupath_project,
                download_progress_callable=update_progress,
            )

            main_download_progress_ui.update(main_task, completed=100, total=100)

        message = f"Downloaded results of run '{run_id}' to '{destination_directory}'"
        logger.info(message)
        console.print(message, style="info")
    except NotFoundException as e:
        logger.warning("Run with ID '%s' not found: %s", run_id, e)
        console.print(f"[warning]Warning:[/warning] Run with ID '{run_id}' not found.")
        sys.exit(2)
    except ValueError as e:
        logger.warning("Bad input to download results of run with ID '%s': %s", run_id, e)
        console.print(f"[warning]Warning:[/warning] Bad input to download results of run with ID '{run_id}': {e}")
        sys.exit(2)
    except Exception as e:
        logger.exception("Failed to download results of run with ID '%s'", run_id)
        console.print(
            f"[error]Error:[/error] Failed to download results of run with ID '{run_id}': {type(e).__name__}: {e}"
        )
        sys.exit(1)


# TODO(Helmut): Implement result delete when available in platform
@result_app.command("delete")
def result_delete() -> None:
    """Delete results of a run."""
    console.print(MESSAGE_NOT_YET_IMPLEMENTED)
    sys.exit(1)



"""Applications resource module for the Aignostics platform.

This module provides classes for interacting with application resources in the Aignostics API.
It includes functionality for listing applications and managing application versions.
"""

import builtins
import re
import typing as t
from operator import itemgetter

from aignx.codegen.api.public_api import PublicApi
from aignx.codegen.models import ApplicationReadResponse as Application
from aignx.codegen.models import ApplicationVersionReadResponse as ApplicationVersion

from aignostics.platform.resources.utils import paginate


class Versions:
    """Resource class for managing application versions.

    Provides operations to list and retrieve application versions.
    """

    APPLICATION_VERSION_REGEX = re.compile(r"(?P<application_id>[^:]+):v?(?P<version>[^:]+)")

    def __init__(self, api: PublicApi) -> None:
        """Initializes the Versions resource with the API platform.

        Args:
            api (PublicApi): The configured API platform.
        """
        self._api = api

    def list(self, application: Application | str) -> t.Iterator[ApplicationVersion]:
        """Find all versions for a specific application.

        Args:
            application (Application | str): The application to find versions for, either object or id

        Returns:
            Iterator[ApplicationVersion]: A Iterator over the available application versions.

        Raises:
            Exception: If the API request fails.
        """
        application_id = application.application_id if isinstance(application, Application) else application

        return paginate(
            self._api.list_versions_by_application_id_v1_applications_application_id_versions_get,
            application_id=application_id,
        )

    # TODO(Andreas,Helmut): Discuss. This is getting an application version and returning it.
    # Can we make this a find_by_id, just getting an application_version_id as a string,
    # and returning the application version object?
    def details(self, application_version: ApplicationVersion | str) -> ApplicationVersion:
        """Retrieves details for a specific application version.

        Args:
            application_version (ApplicationVersion | str): The ID of the application version.

        Returns:
            ApplicationVersion: The version details.

        Raises:
            RuntimeError: If the application version ID is invalid or if the API request fails.
            Exception: If the API request fails.
        """
        if isinstance(application_version, ApplicationVersion):
            application_id = application_version.application_id
            version = application_version.version
        else:
            # split by colon
            m = Versions.APPLICATION_VERSION_REGEX.match(application_version)
            if not m:
                msg = f"Invalid application_version_id: {application_version}"
                raise RuntimeError(msg)
            application_id = m.group("application_id")
            version = m.group("version")

        application_versions = self._api.list_versions_by_application_id_v1_applications_application_id_versions_get(
            application_id=application_id,
            version=version,
        )
        if len(application_versions) != 1:
            # this invariance is enforced by the system. If that error occurs, we have an internal error
            msg = "Internal server error. Please contact Aignostics support."
            raise RuntimeError(msg)
        return application_versions[0]

    # TODO(Andreas): Remove when supported in backend
    def list_sorted(self, application: Application | str) -> builtins.list[ApplicationVersion]:
        """Get application versions sorted by semver, descending.

        Args:
            application (Application | str): The application to find versions for, either object or id

        Returns:
            list[ApplicationVersion]: List of version objects sorted by semantic versioning (latest first),
                or empty list if no versions are found
        """
        versions = builtins.list(self.list(application=application))

        # If no versions available
        if not versions:
            return []

        # Extract semantic versions from the version property
        versions_with_semver = []
        for v in versions:
            try:
                # Split into major, minor, patch components for proper comparison
                version_parts = [int(x) for x in v.version.split(".")]
                versions_with_semver.append((v, version_parts))
            except (ValueError, AttributeError):
                # If we can't parse the version or version attribute doesn't exist, skip it
                continue

        # Sort by semantic version (major, minor, patch)
        if versions_with_semver:
            versions_with_semver.sort(key=itemgetter(1), reverse=True)
            # Return just the version objects, not the tuples
            return [item[0] for item in versions_with_semver]

        # If we couldn't parse any versions, return all versions as is
        return versions

    def latest(self, application: Application | str) -> ApplicationVersion | None:
        """Get latest version.

        Args:
            application (Application | str): The application to find versions for, either object or id

        Returns:
            ApplicationVersion | None: The latest version id, or None if no versions found.
        """
        sorted_versions = self.list_sorted(application=application)
        return sorted_versions[0] if sorted_versions else None


class Applications:
    """Resource class for managing applications.

    Provides operations to list applications and access version resources.
    """

    def __init__(self, api: PublicApi) -> None:
        """Initializes the Applications resource with the API platform.

        Args:
            api (PublicApi): The configured API platform.
        """
        self._api = api
        self.versions: Versions = Versions(self._api)

    def list(self) -> t.Iterator[Application]:
        """Find all available applications.

        Returns:
            Iterator[Application]: A Iterator over the available applications.

        Raises:
            Exception: If the API request fails.
        """
        return paginate(self._api.list_applications_v1_applications_get)



"""Handler for wsi files using OpenSlide."""

import xml.etree.ElementTree as ET  # noqa: S405
from pathlib import Path
from typing import Any

import openslide
from openslide import ImageSlide, OpenSlide, open_slide
from PIL.Image import Image

TIFF_IMAGE_DESCRIPTION = "tiff.ImageDescription"


class OpenSlideHandler:
    """Handler for WSI files using OpenSlide."""

    def __init__(self, path: str) -> None:
        self.path = Path(path)
        self.slide: OpenSlide | ImageSlide = open_slide(str(path))

    @classmethod
    def from_file(cls, path: str | Path) -> "OpenSlideHandler":
        """Create an OpenSlideHandler from a file path.

        Args:
            path (str | Path): The path to the WSI file.

        Returns:
            OpenSlideHandler: An instance of OpenSlideHandler initialized with the given path.

        Raises:
            OpenSlideError: If the file cannot be opened as a WSI.
        """
        return cls(str(path))

    def _detect_format(self) -> str | None:
        """Enhanced format detection.

        Returns:
            str: The detected format of the TIFF file.
        """
        props = dict(self.slide.properties)

        # Check for libvips signature in XML metadata
        if TIFF_IMAGE_DESCRIPTION in props:
            try:
                root = ET.fromstring(props[TIFF_IMAGE_DESCRIPTION])  # noqa: S314
                if root.get("xmlns") == "http://www.vips.ecs.soton.ac.uk//dzsave":
                    return "pyramidal-tiff (libvips)"
            except ET.ParseError:
                pass

        # Additional format checks could go here
        # For example, check for BigTIFF indicators
        # BigTIFF uses 64-bit offsets instead of 32-bit
        # This would require reading the TIFF header directly
        base_format = self.slide.detect_format(self.path)
        if base_format == "generic-tiff":
            if self.slide.level_count > 1:
                return "pyramidal-tiff"
            return "tiff"

        return base_format

    def get_thumbnail(self) -> Image:
        """Get thumbnail of the slide.

        Returns:
            Image: Thumbnail image of the slide.

        Raises:
            OpenSlideError: If the slide cannot be opened or if thumbnail generation fails.
        """
        return self.slide.get_thumbnail((256, 256))

    def _parse_xml_image_description(self, xml_string: str) -> dict[str, Any]:  # noqa: C901, PLR6301
        """Parse the XML image description.

        Args:
            xml_string: XML string containing image description.

        Returns:
            dict[str, Any]: Parsed image description as a dictionary with metadata properties.
        """
        try:
            root = ET.fromstring(xml_string)  # noqa: S314
            namespace = {"ns": "http://www.vips.ecs.soton.ac.uk//dzsave"}
            image_desc: dict[str, Any] = {
                "date": root.get("date"),
                "version": root.get("version"),
                "properties": {},
            }
            for prop in root.findall(".//ns:property", namespace):
                name_elem = prop.find("ns:name", namespace)
                if name_elem is None:
                    continue
                name = name_elem.text
                if name is None:
                    continue
                value_elem = prop.find("ns:value", namespace)
                if value_elem is None:
                    continue
                value = value_elem.text
                if value is None:
                    continue
                value_type = value_elem.get("type", "")

                if value_type == "gint":
                    value = int(value)  # type: ignore[assignment]
                elif value_type == "gdouble":
                    value = float(value)  # type: ignore[assignment]
                elif value_type == "VipsRefString":
                    # Handle special libvips string properties
                    if name == "aix-libVips-version":
                        image_desc["libvips_version"] = value
                    elif name == "aix-original-format":
                        image_desc["original_format"] = value

                image_desc["properties"][name] = value

            return image_desc
        except ET.ParseError:
            return {}

    def _get_level_info(self) -> list[dict[str, Any]]:
        """Get detailed information for each level.

        Returns:
            list[dict[str, Any]]: A list of dictionaries containing detailed information for each level
                of the pyramidal image.
        """
        levels = []
        props = dict(self.slide.properties)
        base_mpp_x = float(props.get(openslide.PROPERTY_NAME_MPP_X, 0))
        base_mpp_y = float(props.get(openslide.PROPERTY_NAME_MPP_Y, 0))

        for level in range(self.slide.level_count):
            width, height = self.slide.level_dimensions[level]
            downsample = self.slide.level_downsamples[level]

            tile_width = int(props.get(f"openslide.level[{level}].tile-width", 256))
            tile_height = int(props.get(f"openslide.level[{level}].tile-height", 256))

            # Calculate number of tiles
            tiles_x = (width + tile_width - 1) // tile_width
            tiles_y = (height + tile_height - 1) // tile_height

            level_info = {
                "index": level,
                "dimensions": {
                    "width": width,
                    "height": height,
                    "total_pixels": width * height,
                    "aspect_ratio": width / height if height else 0,
                },
                "downsample": downsample,
                "resolution": {
                    "mpp_x": base_mpp_x * downsample if base_mpp_x else 0,
                    "mpp_y": base_mpp_y * downsample if base_mpp_y else 0,
                },
                "tile": {
                    "width": tile_width,
                    "height": tile_height,
                    "grid": {"x": tiles_x, "y": tiles_y, "total": tiles_x * tiles_y},
                },
            }
            levels.append(level_info)

        return levels

    def get_metadata(self) -> dict[str, Any]:
        """Get comprehensive slide metadata.

        Returns:
            dict[str, Any]: A dictionary containing detailed metadata about the image,
                including format, file information, dimensions, resolution, levels, and other properties.
        """
        props = dict(self.slide.properties)
        file_size = self.path.stat().st_size
        base_width, base_height = self.slide.dimensions

        metadata = {
            "format": self._detect_format(),
            "file": {
                "path": str(self.path),
                "size": file_size,
                "size_human": f"{file_size / (1024 * 1024 * 1024):.2f} GB",
            },
            "dimensions": {"width": base_width, "height": base_height},
            "resolution": {
                "mpp_x": float(props.get(openslide.PROPERTY_NAME_MPP_X, 0)),
                "mpp_y": float(props.get(openslide.PROPERTY_NAME_MPP_Y, 0)),
                "unit": props.get("tiff.ResolutionUnit", "unknown"),
                "x_resolution": float(props.get("tiff.XResolution", 0)),
                "y_resolution": float(props.get("tiff.YResolution", 0)),
            },
            "bounds": {
                "x": int(props.get(openslide.PROPERTY_NAME_BOUNDS_X, 0)),
                "y": int(props.get(openslide.PROPERTY_NAME_BOUNDS_X, 0)),
                "width": int(props.get(openslide.PROPERTY_NAME_BOUNDS_X, base_width)),
                "height": int(props.get(openslide.PROPERTY_NAME_BOUNDS_X, base_height)),
            },
            "tile": {
                "width": int(props.get("openslide.level[0].tile-width", 256)),
                "height": int(props.get("openslide.level[0].tile-height", 256)),
            },
            "levels": {"count": self.slide.level_count, "data": self._get_level_info()},
            "properties": {},
        }

        # Parse image description if available
        if TIFF_IMAGE_DESCRIPTION in props:
            image_desc = self._parse_xml_image_description(props[TIFF_IMAGE_DESCRIPTION])
            if image_desc:
                # Initialize "properties.image" as a dict if needed
                if "properties" not in metadata:
                    metadata["properties"] = {}
                metadata["properties"]["image"] = image_desc  # type: ignore[index, assignment]
                if "libvips_version" in image_desc:
                    metadata["generator"] = f"libvips {image_desc['libvips_version']}"

        # Include vendor information
        if "openslide.vendor" in props:
            metadata["vendor"] = props["openslide.vendor"]

        # Add associated images if any
        associated_images = list(self.slide.associated_images.keys())
        if associated_images:
            metadata["associated_images"] = associated_images  # type: ignore[assignment]

        return metadata

    def close(self) -> None:
        """Close the OpenSlide object."""
        self.slide.close()

    def __enter__(self) -> "OpenSlideHandler":
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: object,
    ) -> None:
        """Exit the context manager.

        Args:
            exc_type: The exception type if an exception was raised.
            exc_val: The exception value if an exception was raised.
            exc_tb: The traceback if an exception was raised.
        """
        self.close()



"""
Utility functions for the Aignostics client resources.

This module provides helper functions for working with the Aignostics API, including
pagination utilities to handle API responses that span multiple pages. These utility
functions are designed to be used internally by the SDK's resource classes.
"""

from collections.abc import Callable, Iterator
from typing import TypeVar

from aignx.codegen.exceptions import NotFoundException

T = TypeVar("T")

PAGE_SIZE = 20


def paginate(func: Callable[..., list[T]], *args: object, page_size: int = PAGE_SIZE, **kwargs: object) -> Iterator[T]:
    """
    A generator function that handles pagination for API calls.

    This function takes a callable that returns a list of items and handles pagination
    by repeatedly calling the function with increasing page numbers until either
    a page returns fewer items than the requested page size or a NotFoundException is raised.

    Args:
        func (Callable[..., list[T]]): The function to paginate, which should accept page and page_size parameters.
        *args: Positional arguments to pass to the function.
        page_size (int): The number of items to request per page
        **kwargs: Keyword arguments to pass to the function.

    Yields:
        Individual items from all pages.

    Example:
        >>> def list_items(page=1, page_size=20):
        ...     # API call that returns a list of items for the given page
        ...     return [f"item_{i}" for i in range(page_size)]
        >>> items = list(paginate(list_items))
        >>> print(len(items))
    """
    page = 1
    while True:
        try:
            results = func(*args, page=page, page_size=page_size, **kwargs)
            if not results:
                break
            yield from results
            if len(results) < page_size:
                break
            page += 1
        except NotFoundException:
            break  # We've paginated beyond the last page



"""Application describe page, including submission stepper."""

import sys
import time
from multiprocessing import Manager
from pathlib import Path
from typing import Any

from nicegui import app, binding, ui  # noq
from nicegui import run as nicegui_run

from aignostics.utils import GUILocalFilePicker, get_logger, get_user_data_directory

from .._service import Service  # noqa: TID252
from .._utils import get_mime_type_for_artifact  # noqa: TID252
from ._frame import _frame
from ._utils import (
    application_id_to_icon,
    mime_type_to_icon,
)

logger = get_logger(__name__)

WIDTH_100 = "width: 100%"
WIDTH_1200px = "width: 1200px; max-width: none"
MESSAGE_METADATA_GRID_IS_NOT_INITIALIZED = "Metadata grid is not initialized."


@binding.bindable_dataclass
class SubmitForm:
    """Submit form."""

    application_version_id: str | None = None
    source: Path | None = None
    wsi_step_label: ui.label | None = None
    wsi_next_button: ui.button | None = None
    wsi_spinner: ui.spinner | None = None
    metadata: list[dict[str, Any]] | None = None
    metadata_grid: ui.aggrid | None = None
    metadata_exclude_button: ui.button | None = None
    metadata_next_button: ui.button | None = None
    submission_upload_button: ui.button | None = None
    submission_submit_button: ui.button | None = None


submit_form = SubmitForm()

upload_message_queue = Manager().Queue()

service = Service()


async def _page_application_describe(application_id: str) -> None:  # noqa: C901, PLR0912, PLR0915
    """Describe Application.

    Args:
        application_id (str): The application ID.
    """
    spinner = ui.spinner(size="xl").classes("fixed inset-0 m-auto")
    application = await nicegui_run.io_bound(service.application, application_id)
    spinner.set_visibility(False)

    if application is None:
        await _frame(
            navigation_title=f"{application_id}",
            navigation_icon="bug_report",
            navigation_icon_color="negative",
            navigation_icon_tooltip="Could not load application",
            left_sidebar=True,
            args={"application_id": application_id},
        )
        ui.label(f"Failed to get application '{application_id}'").mark("LABEL_ERROR")
        return

    await _frame(
        navigation_title=f"{application.name if application else ''}",
        navigation_icon=application_id_to_icon(application_id),
        navigation_icon_color="primary",
        navigation_icon_tooltip=application_id,
        left_sidebar=True,
        args={"application_id": application_id},
    )

    application_versions = service.application_versions(application)
    latest_application_version = application_versions[0]
    latest_application_version_id = latest_application_version.application_version_id
    submit_form.application_version_id = latest_application_version_id

    with ui.dialog() as release_notes_dialog, ui.card().style(WIDTH_1200px):
        ui.label(f"Release notes of {application.name}").classes("text-h5")
        with ui.scroll_area().classes("w-full h-100"):
            for application_version in application_versions:
                ui.label(f"Version {application_version.version}").classes("text-h6")
                ui.markdown(application_version.changelog)
        with ui.row(align_items="end").classes("w-full"), ui.column(align_items="end").classes("w-full"):
            ui.button("Close", on_click=release_notes_dialog.close)

    with ui.row(align_items="start").classes("justify-center w-full"):
        with ui.column(), ui.expansion(application.name, icon="info").classes("full-width") as application_info:
            ui.markdown(application.description.replace("\n", "\n\n"))
        ui.space()
        with ui.row(align_items="center"):
            with ui.button("Release Notes", icon="change_history", on_click=release_notes_dialog.open):
                ui.tooltip("Show change notes of this application.")
            for regulatory_class in application.regulatory_classes:
                if regulatory_class in {"RUO", "RuO"}:
                    with ui.link(
                        target="https://www.fda.gov/regulatory-information/search-fda-guidance-documents/distribution-in-vitro-diagnostic-products-labeled-research-use-only-or-investigational-use-only",
                        new_tab=True,
                    ):
                        ui.image("/application_assets/ruo.png").style("width: 70px; height: 36px")
                        ui.tooltip("Go to explanation of this regulatory class")

                elif regulatory_class == "demo":
                    with ui.icon("network_check", size="lg", color="orange"):
                        ui.tooltip("For testing only.")
                else:
                    ui.label(f"{regulatory_class}")
            if not application.regulatory_classes:
                with ui.link(
                    target="https://www.fda.gov/regulatory-information/search-fda-guidance-documents/distribution-in-vitro-diagnostic-products-labeled-research-use-only-or-investigational-use-only",
                    new_tab=True,
                ):
                    ui.image("/application_assets/ruo.png").style("width: 70px; height: 36px")

    async def _select_source(data: bool = False) -> None:
        """Open a file picker dialog and show notifier when closed again."""
        from nicegui import ui  # noqa: PLC0415

        result = await GUILocalFilePicker(
            str(get_user_data_directory("datasets") if data else Path.home()), multiple=False
        )  # type: ignore
        if result and len(result) > 0:
            path = Path(result[0])
            if not path.is_dir():
                submit_form.source = None
                submit_form.wsi_step_label.set_text(
                    "Select a folder with whole slide images you want to analyze"
                ) if submit_form.wsi_step_label else None
                submit_form.wsi_next_button.disable() if submit_form.wsi_next_button else None
                ui.notify("The selected path is not a directory. Please select a valid directory.", type="warning")
            else:
                submit_form.source = path
                submit_form.wsi_step_label.set_text(
                    f"Selected folder {submit_form.source} to analyze."
                ) if submit_form.wsi_step_label else None
                submit_form.wsi_next_button.enable() if submit_form.wsi_next_button else None
                ui.notify(
                    f"You chose directory {submit_form.source}. Feel free to continue to the next step.",
                    type="positive",
                )
        else:
            submit_form.source = None
            submit_form.wsi_step_label.set_text(
                "Select a folder with whole slide images you want to analyze"
            ) if submit_form.wsi_step_label else None
            submit_form.wsi_next_button.disable() if submit_form.wsi_next_button else None
            ui.notify(
                "You did not make a selection. You must choose a source directory to upload from.",
                type="warning",
            )

    async def _pytest_home() -> None:  # noqa: RUF029
        """Select home folder."""
        from nicegui import ui  # noqa: PLC0415

        submit_form.source = Path.home()
        submit_form.wsi_step_label.set_text(
            f"Selected folder {submit_form.source} to analyze."
        ) if submit_form.wsi_step_label else None
        submit_form.wsi_next_button.enable() if submit_form.wsi_next_button else None
        ui.notify(
            f"You chose directory {submit_form.source}. Feel free to continue to the next step.",
            type="positive",
        )

    async def _on_wsi_next_click() -> None:
        """Handle the 'Next' button click in WSI step.

        This function:
        1. Generates metadata from the selected source directory
        2. Updates the metadata grid with the generated data
        3. Moves to the next step

        Raises:
            RuntimeError: If the metadata grid is not initialized or if the generated metadata is None.
        """
        if submit_form.source and submit_form.metadata_grid and submit_form.wsi_spinner and submit_form.wsi_next_button:
            try:
                ui.notify(f"Finding WSIs and generating metadata for {submit_form.source}...", type="info")
                if submit_form.metadata_grid is None:
                    logger.error(MESSAGE_METADATA_GRID_IS_NOT_INITIALIZED)  # type: ignore[unreachable]
                    return
                submit_form.wsi_spinner.set_visibility(True)
                submit_form.wsi_next_button.set_visibility(False)
                submit_form.metadata_grid.options["rowData"] = await nicegui_run.cpu_bound(
                    Service.generate_metadata_from_source_directory,
                    str(submit_form.application_version_id),
                    submit_form.source,
                    True,
                    [".*:staining_method=H&E"],
                )
                if submit_form.metadata_grid.options["rowData"] is None:
                    msg = "nicegui_run.cpu_bound(Service.generate_metadata_from_source_directory) returned None"
                    logger.error(msg)
                    submit_form.wsi_next_button.set_visibility(True)
                    submit_form.wsi_spinner.set_visibility(False)
                    raise RuntimeError(msg)  # noqa: TRY301
                submit_form.wsi_next_button.set_visibility(True)
                submit_form.wsi_spinner.set_visibility(False)
                submit_form.metadata_grid.update()
                ui.notify(
                    f"Found {len(submit_form.metadata_grid.options['rowData'])} slides for analysis."
                    "Please provide missing metadata.",
                    type="positive",
                )
                stepper.next()
            except Exception as e:
                logger.exception("Error generating metadata from source directory")
                ui.notify(f"Error generating metadata: {e!s}", type="warning")
                raise
        else:
            ui.notify("No source directory selected", type="warning")

    with ui.dialog() as info_dialog, ui.card().style("width: 1200px; max-width: none; height: 1000px"):  # noqa: PLR1702
        if submit_form.application_version_id is None:
            return
        with ui.scroll_area().classes("w-full h-[calc(100vh-2rem)]"):
            for application_version in application_versions:
                if application_version.application_version_id == submit_form.application_version_id:
                    ui.label(f"Latest changes in v{application_version.version}").classes("text-h5")
                    ui.markdown(application_version.changelog)
                    ui.label("Expected Input Artifacts:").classes("text-h5")
                    for artifact in application_version.input_artifacts:
                        with ui.expansion(
                            artifact.name, icon=mime_type_to_icon(get_mime_type_for_artifact(artifact))
                        ).classes("w-full"):
                            ui.label("Metadata")
                            ui.json_editor({
                                "content": {"json": artifact.metadata_schema},
                                "mode": "tree",
                                "readOnly": True,
                                "mainMenuBar": False,
                                "navigationBar": True,
                                "statusBar": False,
                            }).style(WIDTH_100)
                    ui.label("Generated output artifacts:").classes("text-h5")
                    for artifact in application_version.output_artifacts:
                        with ui.expansion(
                            artifact.name, icon=mime_type_to_icon(get_mime_type_for_artifact(artifact))
                        ).classes("w-full"):
                            ui.label(f"Scope: {artifact.scope}")
                            ui.label(f"Mime Type: {artifact.mime_type}")
                            ui.label("Metadata")
                            ui.json_editor({
                                "content": {"json": artifact.metadata_schema},
                                "mode": "tree",
                                "readOnly": True,
                                "mainMenuBar": False,
                                "navigationBar": True,
                                "statusBar": False,
                            }).style(WIDTH_100)
                    break
        with ui.row(align_items="end").classes("w-full"), ui.column(align_items="end").classes("w-full"):
            ui.button("Close", on_click=info_dialog.close)

    with ui.stepper().props("vertical").classes("w-full") as stepper:  # noqa: PLR1702
        with ui.step("Select Application Version"):
            with ui.row().classes("w-full justify-center"):
                with ui.column():
                    ui.label(
                        f"Select the version of {application.name} you want to run. Not sure? "
                        "Click “Next” to auto-select the latest version"
                    )
                    ui.select(
                        {version.application_version_id: version.version for version in application_versions},
                        value=latest_application_version_id,
                    ).bind_value(submit_form, "application_version_id")
                ui.space()
                with ui.column(), ui.button(icon="info", on_click=info_dialog.open):
                    ui.tooltip(
                        "Show technical information about expected input and output of this application version."
                    )
            with ui.stepper_navigation():
                ui.button("Next", on_click=lambda: (application_info.close(), stepper.next())).mark(  # type: ignore[func-returns-value]
                    "BUTTON_APPLICATION_VERSION_NEXT"
                )

        with ui.step("Select Whole Slide Images"):
            submit_form.wsi_step_label = ui.label(
                "Select the folder with the whole slide images you want to analyze then click Next."
            )
            with ui.stepper_navigation():
                if "pytest" in sys.modules:
                    ui.button("Home", on_click=_pytest_home, icon="folder").mark("BUTTON_PYTEST_HOME")
                with ui.button(
                    "Data", on_click=lambda _: _select_source(True), icon="folder_special", color="purple-400"
                ).mark("BUTTON_WSI_SELECT_DATA"):
                    ui.tooltip("Select folder within Launchpad datasets directory")
                with ui.button("Custom", on_click=_select_source, icon="folder").mark("BUTTON_WSI_SELECT_CUSTOM"):
                    ui.tooltip("Select custom folder starting at your home directory")
                submit_form.wsi_next_button = ui.button("Next", on_click=_on_wsi_next_click)
                submit_form.wsi_next_button.mark("BUTTON_WSI_NEXT").disable()
                submit_form.wsi_spinner = ui.spinner(size="lg")
                submit_form.wsi_spinner.set_visibility(False)
                ui.button("Back", on_click=stepper.previous).props("flat")

        with ui.step("Choose Images and Edit Metadata"):
            ui.markdown(
                """
                The Launchpad has found all compatible slide files in your selected folder.

                1. Check the slides that have been found.
                    If you wish to exclude any from analysis at this point, check the boxes next to those slides
                    and click “Exclude” to remove them.
                2. For the slides you wish to analyze,
                    provide the missing metadata to finalize the upload.
                    Double click red cells to edit the missing data with the available options.
                3. Once all the metadata has been added and your slide selection has been finalized,
                    click “Next”.

                You can revert to the previous step and reupload at any point by clicking “Back”.
                """
            )

            async def _pytest_meta() -> None:  # noqa: RUF029
                if submit_form.metadata_grid is None:
                    logger.error(MESSAGE_METADATA_GRID_IS_NOT_INITIALIZED)
                    return
                if submit_form.metadata_next_button is None:
                    logger.error("Metadata next button is not initialized.")
                    return
                submit_form.metadata_next_button.enable()
                ui.notify("Your metadata is now valid! Feel free to continue to the next step.", type="positive")

            async def _validate() -> None:
                if submit_form.metadata_grid is None:
                    logger.error(MESSAGE_METADATA_GRID_IS_NOT_INITIALIZED)
                    return
                rows = await submit_form.metadata_grid.get_client_data()
                valid = True
                for row in rows:
                    if (
                        row["tissue"]
                        not in {
                            "ADRENAL_GLAND",
                            "BLADDER",
                            "BONE",
                            "BRAIN",
                            "BREAST",
                            "COLON",
                            "LIVER",
                            "LUNG",
                            "LYMPH_NODE",
                        }
                    ) or (
                        row["disease"]
                        not in {
                            "BREAST_CANCER",
                            "BLADDER_CANCER",
                            "COLORECTAL_CANCER",
                            "LIVER_CANCER",
                            "LUNG_CANCER",
                        }
                    ):
                        valid = False
                        break
                if submit_form.metadata_next_button is None:
                    logger.error("Metadata next button is not initialized.")
                    return
                if not valid:
                    submit_form.metadata_next_button.disable()
                else:
                    submit_form.metadata_next_button.enable()
                    ui.notify("Your metadata is now valid. Feel free to continue to the next step.", type="positive")
                submit_form.metadata_grid.run_grid_method("autoSizeAllColumns")

            async def _metadata_next() -> None:
                if submit_form.metadata_grid is None or submit_form.submission_upload_button is None:
                    logger.error(MESSAGE_METADATA_GRID_IS_NOT_INITIALIZED)
                    return
                if "pytest" in sys.modules:
                    rows = submit_form.metadata_grid.options["rowData"]
                    for row in rows:
                        row["tissue"] = "LUNG"
                        row["disease"] = "LUNG_CANCER"
                    submit_form.metadata = rows
                else:
                    submit_form.metadata = await submit_form.metadata_grid.get_client_data()
                _upload_ui.refresh(submit_form.metadata)
                submit_form.submission_upload_button.enable()
                if "pytest" in sys.modules:
                    message = f"Prepared upload UI with metadata '{submit_form.metadata}' for pytest."
                    logger.debug(message)
                    ui.notify("Prepared upload UI.", type="info")
                stepper.next()

            async def _delete_selected() -> None:
                if submit_form.metadata_grid is None or submit_form.metadata_exclude_button is None:
                    logger.error(MESSAGE_METADATA_GRID_IS_NOT_INITIALIZED)
                    return
                selected_rows = await submit_form.metadata_grid.get_selected_rows()
                if (selected_rows is None) or (len(selected_rows) == 0):
                    return
                submit_form.metadata = await submit_form.metadata_grid.get_client_data()
                submit_form.metadata[:] = [row for row in submit_form.metadata if row not in selected_rows]
                submit_form.metadata_grid.options["rowData"] = submit_form.metadata
                submit_form.metadata_grid.update()
                submit_form.metadata_exclude_button.set_text("Exclude")
                submit_form.metadata_exclude_button.disable()

            async def _handle_grid_selection_changed() -> None:
                if submit_form.metadata_grid is None or submit_form.metadata_exclude_button is None:
                    logger.error("Metadata grid or button is not initialized.")
                    return
                rows = await submit_form.metadata_grid.get_selected_rows()
                if rows:
                    submit_form.metadata_exclude_button.set_text(f"Exclude {len(rows)} slides")
                    submit_form.metadata_exclude_button.enable()
                else:
                    submit_form.metadata_exclude_button.set_text("Exclude")
                    submit_form.metadata_exclude_button.disable()

            thumbnail_renderer_js = """
                class ThumbnailRenderer {
                    init(params) {
                        this.eGui = document.createElement('img');
                        this.eGui.setAttribute('src', `/thumbnail?source=${encodeURIComponent(params.data.source)}`);
                        this.eGui.setAttribute('style', 'height:70px; width: 70px');
                        this.eGui.setAttribute('alt', `${params.data.reference}`);
                    }
                    getGui() {
                        return this.eGui;
                    }
                }
            """

            submit_form.metadata_grid = (
                ui.aggrid({
                    "columnDefs": [
                        {"headerName": "Reference", "field": "reference_short", "checkboxSelection": True},
                        {
                            "headerName": "Thumbnail",
                            "field": "thumbnail",
                            ":cellRenderer": thumbnail_renderer_js,
                            "autoHeight": True,
                        },
                        {
                            "headerName": "Tissue",
                            "field": "tissue",
                            "editable": True,
                            "cellEditor": "agSelectCellEditor",
                            "cellEditorParams": {
                                "values": [
                                    "ADRENAL_GLAND",
                                    "BLADDER",
                                    "BONE",
                                    "BRAIN",
                                    "BREAST",
                                    "COLON",
                                    "LIVER",
                                    "LUNG",
                                    "LYMPH_NODE",
                                ],
                                "valueListGap": 10,
                            },
                            "cellClassRules": {
                                "bg-red-300": "!new Set(['ADRENAL_GLAND', 'BLADDER', 'BONE', 'BRAIN',"
                                "'BREAST', 'COLON', 'LIVER', 'LUNG', 'LYMPH_NODE']).has(x)",
                                "bg-green-300": "new Set(['ADRENAL_GLAND', 'BLADDER', 'BONE', 'BRAIN',"
                                "'BREAST', 'COLON', 'LIVER', 'LUNG', 'LYMPH_NODE']).has(x)",
                            },
                        },
                        {
                            "headerName": "Disease",
                            "field": "disease",
                            "editable": True,
                            "cellEditor": "agSelectCellEditor",
                            "cellEditorParams": {
                                "values": [
                                    "BREAST_CANCER",
                                    "BLADDER_CANCER",
                                    "COLORECTAL_CANCER",
                                    "LIVER_CANCER",
                                    "LUNG_CANCER",
                                ],
                                "valueListGap": 10,
                            },
                            "cellClassRules": {
                                "bg-red-300": "!new Set(['BREAST_CANCER', 'BLADDER_CANCER', "
                                "'COLORECTAL_CANCER', 'LIVER_CANCER', 'LUNG_CANCER']).has(x)",
                                "bg-green-300": "new Set(['BREAST_CANCER', 'BLADDER_CANCER', "
                                "'COLORECTAL_CANCER', 'LIVER_CANCER', 'LUNG_CANCER']).has(x)",
                            },
                        },
                        {"headerName": "File size", "field": "file_size_human"},
                        {"headerName": "MPP", "field": "resolution_mpp"},
                        {"headerName": "Width", "field": "width_px"},
                        {"headerName": "Height", "field": "height_px"},
                        {"headerName": "Staining", "field": "staining_method"},
                        {"headerName": "Source", "field": "source"},
                        {"headerName": "Checksum", "field": "checksum_base64_crc32c"},
                        {"headerName": "Upload progress", "field": "file_upload_progress", "initialHide": True},
                        {
                            "headerName": "Platform Bucket URL",
                            "field": "platform_bucket_url",
                            "initialHide": True,
                        },
                    ],
                    "rowData": [],
                    "rowSelection": "multiple",
                    "stopEditingWhenCellsLoseFocus": True,
                    "enableCellTextSelection": "true",
                    "autoSizeStrategy": {
                        "type": "fitCellContents",
                        "defaultMinWidth": 10,
                        "columnLimits": [{"colId": "source", "minWidth": 150}],
                    },
                    "domLayout": "normal",
                })
                .style("height: 210px")
                .classes("ag-theme-balham-dark" if app.storage.general.get("dark_mode", False) else "ag-theme-balham")
                .on("cellValueChanged", lambda _: _validate())
                .on("selectionChanged", _handle_grid_selection_changed)
                .mark("GRID_METADATA")
            )
            # use ui timer to update the grid class depending on dark mode, with a frequency of once per second
            ui.timer(
                interval=1,
                callback=lambda: submit_form.metadata_grid.classes(
                    add="ag-theme-balham-dark" if app.storage.general.get("dark_mode", False) else "ag-theme-balham",
                    remove="ag-theme-balham" if app.storage.general.get("dark_mode", False) else "ag-theme-balham-dark",
                )
                if submit_form.metadata_grid
                else None,
            )
            with ui.stepper_navigation():
                if "pytest" in sys.modules:
                    ui.button("Select", on_click=_pytest_meta, icon="folder").mark("BUTTON_PYTEST_META")
                submit_form.metadata_next_button = ui.button("Next", on_click=_metadata_next)
                submit_form.metadata_next_button.mark("BUTTON_METADATA_NEXT").disable()
                with ui.button("Exclude selected", on_click=_delete_selected).mark(
                    "BUTTON_DELETE_SELECTED"
                ) as exclude_button:
                    ui.tooltip("Exclude selected slides from analysis")
                submit_form.metadata_exclude_button = exclude_button
                submit_form.metadata_exclude_button.set_text("Exclude")
                submit_form.metadata_exclude_button.disable()
                ui.button("Back", on_click=stepper.previous).props("flat")

        def _submit() -> None:
            """Submit the application run."""
            ui.notify("Submitting application run ...", type="info")
            try:
                run = service.application_run_submit_from_metadata(
                    str(submit_form.application_version_id),
                    submit_form.metadata or [],
                )
            except Exception as e:  # noqa: BLE001
                ui.notify(f"Failed to submit application run: {e}.", type="warning")
                return
            ui.notify(
                f"Application run submitted with id '{run.application_run_id}'. Navigating to application run ...",
                type="positive",
            )
            ui.navigate.to(f"/application/run/{run.application_run_id}")

        async def _upload() -> None:
            """Upload prepared slides."""
            global upload_message_queue  # noqa: PLW0602
            if submit_form.submission_submit_button is None or submit_form.submission_upload_button is None:
                logger.error("Submission submit button is not initialized.")
                return
            ui.notify("Uploading whole slide images to Aignostics Platform ...", type="info")

            await nicegui_run.cpu_bound(
                Service.application_run_upload,
                str(submit_form.application_version_id),
                submit_form.metadata or [],
                str(time.time() * 1000),
                upload_message_queue,
            )
            ui.notify("Upload to Aignostics Platform completed.", type="positive")
            submit_form.submission_submit_button.enable()
            submit_form.submission_upload_button.disable()
            _submit()

        @ui.refreshable
        def _upload_ui(metadata: list[dict[str, Any]]) -> None:
            """Upload UI."""
            with ui.column(align_items="start"):
                ui.label(f"Upload and submit your {len(metadata)} slide(s) for analysis.")
                upload_complete = True
                for row in metadata or []:
                    upload_complete = upload_complete and row["file_upload_progress"] == 1
                    with ui.row(align_items="center"):
                        with ui.circular_progress(value=row["file_upload_progress"], show_value=False):
                            ui.button(icon="cloud_upload").props("flat round").disable()
                        ui.label(f"{row['source']} ({row['file_size_human']})").classes("w-4/5")

        def _update_upload_progress() -> None:
            """Update the upload progress for each file."""
            global upload_message_queue  # noqa: PLW0602
            if submit_form.metadata is None:
                return
            while not upload_message_queue.empty():
                message = upload_message_queue.get()
                if message and isinstance(message, dict) and "reference" in message:
                    for row in submit_form.metadata:
                        if row["reference"] == message["reference"]:
                            if "file_upload_progress" in message:
                                row["file_upload_progress"] = message["file_upload_progress"]
                                break
                            if "platform_bucket_url" in message:
                                row["platform_bucket_url"] = message["platform_bucket_url"]
                                break
                _upload_ui.refresh(submit_form.metadata)

        with ui.step("Slide Submission"):
            _upload_ui([])
            ui.timer(0.1, callback=_update_upload_progress)

            with ui.stepper_navigation():
                with ui.button(
                    "Upload and Submit",
                    on_click=_upload,
                    icon="check",
                ).mark("BUTTON_SUBMISSION_UPLOAD") as submission_upload_button:
                    ui.tooltip(
                        "Upload selected slides to Aignostics platform bucket, "
                        "and submit application run when uploaded."
                    )
                submit_form.submission_upload_button = submission_upload_button
                submit_form.submission_submit_button = ui.button(
                    "Submit",
                    on_click=_submit,
                    icon="check",
                )
                submit_form.submission_submit_button.set_visibility(False)
                submit_form.submission_submit_button.mark("BUTTON_SUBMISSION_SUBMIT").disable()
                ui.button("Back", on_click=stepper.previous).props("flat")



import os
from urllib.request import getproxies

from aignx.codegen.api.public_api import PublicApi
from aignx.codegen.api_client import ApiClient
from aignx.codegen.configuration import Configuration
from aignx.codegen.exceptions import NotFoundException
from aignx.codegen.models import ApplicationReadResponse as Application

from aignostics.platform._authentication import get_token
from aignostics.platform.resources.applications import Applications, Versions
from aignostics.platform.resources.runs import ApplicationRun, Runs
from aignostics.utils import get_logger

from ._settings import settings

logger = get_logger(__name__)


class Client:
    """Main client for interacting with the Aignostics Platform API.

    Provides access to platform resources like applications, versions, and runs.
    Handles authentication and API client configuration.
    """

    applications: Applications
    runs: Runs
    versions: Versions

    def __init__(self, cache_token: bool = True) -> None:
        """Initializes a client instance with authenticated API access.

        Args:
            cache_token (bool): If True, caches the authentication token.
                Defaults to True.

        Sets up resource accessors for applications, versions, and runs.
        """
        try:
            logger.debug("Initializing client with cache_token=%s", cache_token)
            self._api = Client.get_api_client(cache_token=cache_token)
            self.applications: Applications = Applications(self._api)
            self.runs: Runs = Runs(self._api)
            logger.debug("Client initialized successfully.")
        except Exception:
            logger.exception("Failed to initialize client.")
            raise

    def run(self, application_run_id: str) -> ApplicationRun:
        """Finds a specific run by id.

        Args:
            application_run_id (str): The ID of the application run.

        Returns:
            Run: The run object.
        """
        return ApplicationRun(self._api, application_run_id)

    # TODO(Andreas): Provide a /v1/applications/{application_id} endpoint and use that
    def application(self, application_id: str) -> Application:
        """Finds a specific application by id.

        Args:
            application_id (str): The ID of the application.

        Raises:
            NotFoundException: If the application with the given ID is not found.

        Returns:
            Application: The application object.
        """
        applications = self.applications.list()
        for application in applications:
            if application.application_id == application_id:
                return application
        logger.warning("Application with ID '%s' not found.", application_id)
        raise NotFoundException

    @staticmethod
    def get_api_client(cache_token: bool = True) -> PublicApi:
        """Creates and configures an authenticated API client.

        Args:
            cache_token (bool): If True, caches the authentication token.
                Defaults to True.

        Returns:
            ExternalsApi: Configured API client with authentication token.

        Raises:
            RuntimeError: If authentication fails.
        """
        token = get_token(use_cache=cache_token)
        config = Configuration(
            host=settings().api_root,
            ssl_ca_cert=os.getenv("REQUESTS_CA_BUNDLE"),  # point to .cer file of proxy if defined
        )
        config.proxy = getproxies().get("https")  # use system proxy
        client = ApiClient(
            config,
            header_name="Authorization",
            header_value=f"Bearer {token}",
        )
        return PublicApi(client)



"""Layout and theme of GUI."""

from importlib.util import find_spec

__all__ = []

# advertise PageBuilder to enable auto-discovery
if find_spec("nicegui"):
    from ._frame import HEALTH_UPDATE_INTERVAL, frame
    from ._theme import PageBuilder, theme

    __all__ += ["HEALTH_UPDATE_INTERVAL", "PageBuilder", "frame", "theme"]



"""Logging configuration and utilities."""

import contextlib
import logging as python_logging
import os
from logging import Filter as LogggingFilter
from logging import Handler
from pathlib import Path
from typing import Annotated, Literal

import click
from pydantic import AfterValidator, Field
from pydantic_settings import BaseSettings, SettingsConfigDict
from rich.console import Console
from rich.logging import RichHandler

from ._constants import __env_file__, __is_running_in_read_only_environment__, __project_name__
from ._settings import load_settings


def get_logger(name: str | None) -> python_logging.Logger:
    """
    Get a logger instance with the given name or project name as default.

    Args:
        name(str): The name for the logger. If None, uses project name.

    Returns:
        Logger: Configured logger instance.
    """
    if (name is None) or (name == __project_name__):
        return python_logging.getLogger(__project_name__)
    return python_logging.getLogger(f"{__project_name__}.{name}")


def _validate_file_name(file_name: str | None) -> str | None:
    """Validate the file_name is valid and the file writeable.

    - Checks file_name does not yet exist or is a file
    - If not yet existing, checks it can be created
    - If existing file, checks file is writeable

    Args:
        file_name: The file name of the log file

    Returns:
        str | None: The validated file name

    Raises:
        ValueError: If file name is not valid or the file not writeable
    """
    if file_name is None:
        return file_name

    file_path = Path(file_name)
    if file_path.exists():
        if file_path.is_dir():
            message = f"File name {file_path.absolute()} exists but is a directory"
            raise ValueError(message)
        if not os.access(file_path, os.W_OK):
            if file_path.exists():
                message = f"File {file_path.absolute()} is not writable"
                raise ValueError(message)
            return file_name  # This was a race condition, file was deleted in the meantime
    else:
        try:
            file_path.touch(exist_ok=True)
        except OSError as e:
            message = f"File {file_path.absolute()} cannot be created: {e}"
            raise ValueError(message) from e

        with contextlib.suppress(OSError):  # Parallel execution e.g. in tests can create race
            file_path.unlink()

    return file_name


class LogSettings(BaseSettings):
    """Settings for configuring logging behavior."""

    model_config = SettingsConfigDict(
        env_prefix=f"{__project_name__.upper()}_LOG_",
        extra="ignore",
        env_file=__env_file__,
        env_file_encoding="utf-8",
    )

    level: Annotated[
        Literal["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"],
        Field(description="Logging level", default="INFO"),
    ]
    file_enabled: Annotated[
        bool,
        Field(description="Enable logging to file", default=False),
    ]
    file_name: Annotated[
        str,
        AfterValidator(_validate_file_name),
        Field(
            description="Name of the log file",
            default="/dev/stdout" if __is_running_in_read_only_environment__ else f"{__project_name__}.log",
        ),
    ]
    console_enabled: Annotated[
        bool,
        Field(description="Enable logging to console", default=False),
    ]


class CustomFilter(LogggingFilter):
    """Filter to exclude specific dependencies or their messages from logging."""

    def filter(self, record: python_logging.LogRecord) -> bool:  # noqa: PLR6301
        """
        Filter out log records from specific dependencies.

        Args:
            record: The log record to filter.

        Returns:
            bool: True if the record should be logged, False otherwise.
        """
        excluded_dependencies = {"showinfm"}
        if record.name in excluded_dependencies:
            return False
        return not (record.name == "dotenv.main" and record.getMessage().endswith("key doesn't exist."))


def logging_initialize(log_to_logfire: bool = False) -> None:
    """Initialize logging configuration."""
    log_filter = CustomFilter()

    handlers: list[Handler] = []

    settings = load_settings(LogSettings)

    if settings.file_enabled:
        file_handler = python_logging.FileHandler(settings.file_name)
        file_formatter = python_logging.Formatter(
            fmt="%(asctime)s %(process)d %(levelname)s %(name)s %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )
        file_handler.setFormatter(file_formatter)
        file_handler.addFilter(log_filter)
        handlers.append(file_handler)

    if settings.console_enabled:
        rich_handler = RichHandler(
            console=Console(stderr=True),
            markup=True,
            rich_tracebacks=True,
            tracebacks_suppress=[click],
            show_time=True,
            omit_repeated_times=True,
            show_path=True,
            show_level=True,
            enable_link_path=True,
        )
        rich_handler.addFilter(log_filter)
        handlers.append(rich_handler)

    if log_to_logfire:
        from importlib.util import find_spec  # noqa: PLC0415

        if find_spec("logfire"):
            import logfire  # noqa: PLC0415

            logfire_handler = logfire.LogfireLoggingHandler()
            logfire_handler.addFilter(log_filter)
            handlers.append(logfire_handler)

    if not handlers:
        handlers = [python_logging.NullHandler()]
    python_logging.basicConfig(
        level=settings.level,
        format="%(name)s %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        handlers=handlers,
    )




"""
    Aignostics Platform API reference

     Pagination is done via `page` and `page_size`. Sorting via `sort` query parameter. The `sort` query parameter can be provided multiple times. The sorting direction can be indicated via `+` (ascending) or `-` (descending) (e.g. `/v1/applications?sort=+name)`.

    The version of the OpenAPI document: 1.0.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from pydantic import BaseModel, ConfigDict, StrictStr
from typing import Any, ClassVar, Dict, List
from aignx.codegen.models.payload_input_artifact import PayloadInputArtifact
from aignx.codegen.models.payload_output_artifact import PayloadOutputArtifact
from typing import Optional, Set
from typing_extensions import Self

class PayloadItem(BaseModel):
    """
    PayloadItem
    """ # noqa: E501
    item_id: StrictStr
    input_artifacts: Dict[str, PayloadInputArtifact]
    output_artifacts: Dict[str, PayloadOutputArtifact]
    __properties: ClassVar[List[str]] = ["item_id", "input_artifacts", "output_artifacts"]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of PayloadItem from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([
        ])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of each value in input_artifacts (dict)
        _field_dict = {}
        if self.input_artifacts:
            for _key_input_artifacts in self.input_artifacts:
                if self.input_artifacts[_key_input_artifacts]:
                    _field_dict[_key_input_artifacts] = self.input_artifacts[_key_input_artifacts].to_dict()
            _dict['input_artifacts'] = _field_dict
        # override the default output from pydantic by calling `to_dict()` of each value in output_artifacts (dict)
        _field_dict = {}
        if self.output_artifacts:
            for _key_output_artifacts in self.output_artifacts:
                if self.output_artifacts[_key_output_artifacts]:
                    _field_dict[_key_output_artifacts] = self.output_artifacts[_key_output_artifacts].to_dict()
            _dict['output_artifacts'] = _field_dict
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of PayloadItem from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "item_id": obj.get("item_id"),
            "input_artifacts": dict(
                (_k, PayloadInputArtifact.from_dict(_v))
                for _k, _v in obj["input_artifacts"].items()
            )
            if obj.get("input_artifacts") is not None
            else None,
            "output_artifacts": dict(
                (_k, PayloadOutputArtifact.from_dict(_v))
                for _k, _v in obj["output_artifacts"].items()
            )
            if obj.get("output_artifacts") is not None
            else None
        })
        return _obj



"""Client specific constants."""

TODO_URL = "https://todo"
API_ROOT_PRODUCTION = "https://platform.aignostics.com"
AUDIENCE_PRODUCTION = "https://aignostics-platform-samia"
AUTHORIZATION_BASE_URL_PRODUCTION = "https://aignostics-platform.eu.auth0.com/authorize"
TOKEN_URL_PRODUCTION = "https://aignostics-platform.eu.auth0.com/oauth/token"  # noqa: S105
REDIRECT_URI_PRODUCTION = "http://localhost:8989/"
DEVICE_URL_PRODUCTION = "https://aignostics-platform.eu.auth0.com/oauth/device/code"
JWS_JSON_URL_PRODUCTION = "https://aignostics-platform.eu.auth0.com/.well-known/jwks.json"

API_ROOT_STAGING = "https://platform-staging.aignostics.com"
# TODO (Andreas): hhva: please fill in
AUDIENCE_STAGING = TODO_URL
AUTHORIZATION_BASE_URL_STAGING = TODO_URL
TOKEN_URL_STAGING = TODO_URL
REDIRECT_URI_STAGING = TODO_URL
DEVICE_URL_STAGING = TODO_URL
JWS_JSON_URL_STAGING = TODO_URL

API_ROOT_DEV = "https://platform-dev.aignostics.com"
AUDIENCE_DEV = "https://dev-8ouohmmrbuh2h4vu-samia"
AUTHORIZATION_BASE_URL_DEV = "https://dev-8ouohmmrbuh2h4vu.eu.auth0.com/authorize"
TOKEN_URL_DEV = "https://dev-8ouohmmrbuh2h4vu.eu.auth0.com/oauth/token"  # noqa: S105
REDIRECT_URI_DEV = "http://localhost:8989/"
DEVICE_URL_DEV = "https://dev-8ouohmmrbuh2h4vu.eu.auth0.com/oauth/device/code"
JWS_JSON_URL_DEV = "https://dev-8ouohmmrbuh2h4vu.eu.auth0.com/.well-known/jwks.json"



"""Health models and status definitions for service health checks."""

from enum import StrEnum
from typing import ClassVar, Self

from pydantic import BaseModel, Field, model_validator


class _HealthStatus(StrEnum):
    UP = "UP"
    DOWN = "DOWN"


class Health(BaseModel):
    """Represents the health status of a service with optional components and failure reasons.

    - A health object can have child components, i.e. health forms a tree.
    - Any node in the tree can set itself to DOWN. In this case the node is required
        to set the reason attribute. If reason is not set when DOWN,
        automatic model validation of the tree will fail.
    - DOWN'ness is propagated to parent health objects. I.e. the health of a parent
        node is automatically set to DOWN if any of its child components are DOWN. The
        child components leading to this will be listed in the reason.
    - The root of the health tree is computed in the system module. The health of other
        modules is automatically picked up by the system module.
    """

    Code: ClassVar[type[_HealthStatus]] = _HealthStatus
    status: _HealthStatus
    reason: str | None = None
    components: dict[str, "Health"] = Field(default_factory=dict)

    def compute_health_from_components(self) -> Self:
        """Recursively compute health status from components.

        - If health is already DOWN, it remains DOWN with its original reason.
        - If health is UP but any component is DOWN, health becomes DOWN with
            a reason listing all failed components.

        Returns:
            Self: The updated health instance with computed status.
        """
        # Skip recomputation if already known to be DOWN
        if self.status == _HealthStatus.DOWN:
            return self

        # No components means we keep the existing status
        if not self.components:
            return self

        # Find all DOWN components
        down_components = []
        for component_name, component in self.components.items():
            # Recursively compute health for each component
            component.compute_health_from_components()
            if component.status == _HealthStatus.DOWN:
                down_components.append(component_name)

        # If any components are DOWN, mark the parent as DOWN
        if down_components:
            self.status = _HealthStatus.DOWN
            if len(down_components) == 1:
                self.reason = f"Component '{down_components[0]}' is DOWN"
            else:
                component_list = "', '".join(down_components)
                self.reason = f"Components '{component_list}' are DOWN"

        return self

    @model_validator(mode="after")
    def validate_health_state(self) -> Self:
        """Validate the health state and ensure consistency.

        - Compute overall health based on component health
        - Ensure UP status has no associated reason
        - Ensure DOWN status always has a reason

        Returns:
            Self: The validated model instance with correct health status.

        Raises:
            ValueError: If validation fails due to inconsistency.
        """
        # First compute health from components
        self.compute_health_from_components()

        # Validate that UP status has no reason
        if (self.status == _HealthStatus.UP) and self.reason:
            msg = f"Health {self.status} must not have reason"
            raise ValueError(msg)

        # Validate that DOWN status always has a reason
        if (self.status == _HealthStatus.DOWN) and not self.reason:
            msg = "Health DOWN must have a reason"
            raise ValueError(msg)

        return self

    def __str__(self) -> str:
        """Return string representation of health status with optional reason for DOWN state.

        Returns:
            str: The health status value, with reason appended if status is DOWN.
        """
        if self.status == _HealthStatus.DOWN and self.reason:
            return f"{self.status.value}: {self.reason}"
        return self.status.value

    def __bool__(self) -> bool:
        """Convert health status to a boolean value.

        Returns:
            bool: True if the status is UP, False otherwise.
        """
        return self.status == _HealthStatus.UP



"""Bucket module."""

from importlib.util import find_spec

__all__ = []

from ._cli import cli
from ._service import Service

__all__ += [
    "Service",
    "cli",
]

# advertise PageBuilder to enable auto-discovery
if find_spec("nicegui"):
    from ._gui import PageBuilder

    __all__ += [
        "PageBuilder",
    ]



"""
This module provides utility functions to support the Aignostics client operations.

It includes helpers for file operations, checksum verification, and Google Cloud Storage
interactions.

These utilities primarily handle file operations, data integrity, and cloud storage
interactions to support the main client functionality.
"""

import base64
import contextlib
import datetime
import re
import tempfile
import typing as t
from collections.abc import Generator
from pathlib import Path
from typing import IO, Any

import google_crc32c
import requests
from tqdm.auto import tqdm

EIGHT_MB = 8_388_608


def mime_type_to_file_ending(mime_type: str) -> str:
    """Converts a MIME type to an appropriate file extension.

    Args:
        mime_type (str): The MIME type string to convert.

    Returns:
        str: The corresponding file extension including the dot.

    Raises:
        ValueError: If the MIME type is not recognized.
    """
    if mime_type == "image/png":
        return ".png"
    if mime_type == "image/tiff":
        return ".tiff"
    if mime_type == "application/vnd.apache.parquet":
        return ".parquet"
    if mime_type in {"application/geo+json", "application/json"}:
        return ".json"
    if mime_type == "text/csv":
        return ".csv"
    msg = f"Unknown mime type: {mime_type}"
    raise ValueError(msg)


# TODO(Andreas,Helmut): discuss
def download_file(signed_url: str, file_path: str, verify_checksum: str) -> None:
    """Downloads a file from a signed URL and verifies its integrity.

    Args:
        signed_url (str): The signed URL to download the file from.
        file_path (str): The local path where the file should be saved.
        verify_checksum (str): The expected CRC32C checksum in base64 encoding.

    Raises:
        ValueError: If the downloaded file's checksum doesn't match the expected value.
        requests.HTTPError: If the download request fails.
    """
    checksum = google_crc32c.Checksum()  # type: ignore[no-untyped-call]
    with requests.get(signed_url, stream=True, timeout=60) as stream:
        stream.raise_for_status()
        with open(file_path, mode="wb") as file:
            total_size = int(stream.headers.get("content-length", 0))
            progress_bar = tqdm(total=total_size, unit="B", unit_scale=True)
            for chunk in stream.iter_content(chunk_size=EIGHT_MB):
                if chunk:
                    file.write(chunk)
                    checksum.update(chunk)  # type: ignore[no-untyped-call]
                    progress_bar.update(len(chunk))
            progress_bar.close()
    downloaded_file = base64.b64encode(checksum.digest()).decode("ascii")  # type: ignore[no-untyped-call]
    if downloaded_file != verify_checksum:
        msg = f"Checksum mismatch: {downloaded_file} != {verify_checksum}"
        raise ValueError(msg)


def generate_signed_url(url: str) -> str:
    """Generates a signed URL for a Google Cloud Storage object.

    Args:
        url (str): The fully qualified bucket URL (e.g. gs://bucket/path/to/object).

    Returns:
        str: A signed URL that can be used to download the object.

    Raises:
        ValueError: If the GS path is invalid or the blob doesn't exist.
    """
    from google.cloud import storage  # noqa: PLC0415, lazy loading for performance

    pattern = r"gs://(?P<bucket_name>[^/]+)/(?P<path>.*)"
    m = re.fullmatch(pattern, url)
    if not m:
        msg = "Invalid google storage URI"
        raise ValueError(msg)
    bucket_name = m.group("bucket_name")
    path = m.group("path")

    storage_client = storage.Client()
    bucket = storage_client.bucket(bucket_name)
    blob = bucket.blob(path)
    if not blob.exists():
        msg = f"Blob does not exist: {url}"
        raise ValueError(msg)

    return t.cast("str", blob.generate_signed_url(expiration=datetime.timedelta(hours=1), method="GET", version="v4"))


def calculate_file_crc32c(file: Path) -> str:
    """Calculates the CRC32C checksum of a file.

    Args:
        file (Path): Path to the file to calculate the checksum for.

    Returns:
        str: The CRC32C checksum in base64 encoding.
    """
    checksum = google_crc32c.Checksum()  # type: ignore[no-untyped-call]
    with open(file, mode="rb") as f:
        for _ in checksum.consume(f, EIGHT_MB):  # type: ignore[no-untyped-call]
            pass
    return base64.b64encode(checksum.digest()).decode("ascii")  # type: ignore[no-untyped-call]


@contextlib.contextmanager
def download_temporarily(signed_url: str, verify_checksum: str) -> Generator[IO[bytes], Any, None]:
    """Downloads a file to a temporary location and provides file handle.

    Args:
        signed_url (str): The signed URL to download the file from.
        verify_checksum (str): The expected CRC32C checksum in base64 encoding.

    Yields:
        IO[bytes]: File handle to the downloaded temporary file.

    Raises:
        ValueError: If the downloaded file's checksum doesn't match the expected value.
        requests.HTTPError: If the download request fails.
    """
    with tempfile.NamedTemporaryFile() as file:
        download_file(signed_url, file.name, verify_checksum)
        yield file



"""Service of the application module."""

import base64
import re
import time
from collections.abc import Callable, Generator
from enum import StrEnum
from http import HTTPStatus
from importlib.util import find_spec
from pathlib import Path
from typing import Any

import google_crc32c
import requests
from pydantic import BaseModel, computed_field

from aignostics.bucket import Service as BucketService
from aignostics.constants import WSI_SUPPORTED_FILE_EXTENSIONS
from aignostics.platform import (
    LIST_APPLICATION_RUNS_MAX_PAGE_SIZE,
    ApiException,
    Application,
    ApplicationRun,
    ApplicationRunData,
    ApplicationRunStatus,
    ApplicationVersion,
    Client,
    InputArtifact,
    InputItem,
    ItemResult,
    ItemStatus,
    NotFoundException,
    OutputArtifactElement,
)
from aignostics.platform import (
    Service as PlatformService,
)
from aignostics.utils import UNHIDE_SENSITIVE_INFO, BaseService, Health, get_logger, sanitize_path_component
from aignostics.wsi import Service as WSIService

from ._settings import Settings
from ._utils import get_file_extension_for_artifact, get_mime_type_for_artifact

has_qupath_extra = find_spec("ijson")
if has_qupath_extra:
    from aignostics.qupath import AddProgress as QuPathAddProgress
    from aignostics.qupath import AnnotateProgress as QuPathAnnotateProgress
    from aignostics.qupath import Service as QuPathService


logger = get_logger(__name__)

APPLICATION_RUN_DOWNLOAD_SLEEP_SECONDS = 5
APPLICATION_RUN_FILE_READ_CHUNK_SIZE = 1024 * 1024 * 1024  # 1GB
APPLICATION_RUN_DOWNLOAD_CHUNK_SIZE = 1024 * 1024  # 1MB
APPLICATION_RUN_UPLOAD_CHUNK_SIZE = 1024 * 1024  # 1MB


class DownloadProgressState(StrEnum):
    """Enum for download progress states."""

    INITIALIZING = "Initializing ..."
    QUPATH_ADD_INPUT = "Adding input slides to QuPath project ..."
    CHECKING = "Checking run status ..."
    WAITING = "Waiting for item completing ..."
    DOWNLOADING = "Downloading artifact ..."
    QUPATH_ADD_RESULTS = "Adding result images to QuPath project ..."
    QUPATH_ANNOTATE_INPUT_WITH_RESULTS = "Annotating input slides in QuPath project with results ..."
    COMPLETED = "Completed."


class DownloadProgress(BaseModel):
    status: DownloadProgressState = DownloadProgressState.INITIALIZING
    run: ApplicationRunData | None = None
    item: ItemResult | None = None
    item_count: int | None = None
    item_index: int | None = None
    item_reference: str | None = None
    artifact: OutputArtifactElement | None = None
    artifact_count: int | None = None
    artifact_index: int | None = None
    artifact_path: Path | None = None
    artifact_download_url: str | None = None
    artifact_size: int | None = None
    artifact_downloaded_chunk_size: int = 0
    artifact_downloaded_size: int = 0
    if has_qupath_extra:
        qupath_add_input_progress: QuPathAddProgress | None = None
        qupath_add_results_progress: QuPathAddProgress | None = None
        qupath_annotate_input_with_results_progress: QuPathAnnotateProgress | None = None

    @computed_field  # type: ignore
    @property
    def total_artifact_count(self) -> int | None:
        if self.item_count and self.artifact_count:
            return self.item_count * self.artifact_count
        return None

    @computed_field  # type: ignore
    @property
    def total_artifact_index(self) -> int | None:
        if self.item_count and self.artifact_count and self.item_index is not None and self.artifact_index is not None:
            return self.item_index * self.artifact_count + self.artifact_index
        return None

    @computed_field  # type: ignore
    @property
    def item_progress_normalized(self) -> float:  # noqa: PLR0911
        """Compute normalized item progress in range 0..1.

        Returns:
            float: The normalized item progress in range 0..1.
        """
        if self.status == DownloadProgressState.DOWNLOADING:
            if (not self.total_artifact_count) or self.total_artifact_index is None:
                return 0.0
            return min(1, float(self.total_artifact_index + 1) / float(self.total_artifact_count))
        if has_qupath_extra:
            if self.status == DownloadProgressState.QUPATH_ADD_INPUT and self.qupath_add_input_progress:
                return self.qupath_add_input_progress.progress_normalized
            if self.status == DownloadProgressState.QUPATH_ADD_RESULTS and self.qupath_add_results_progress:
                return self.qupath_add_results_progress.progress_normalized
            if self.status == DownloadProgressState.QUPATH_ANNOTATE_INPUT_WITH_RESULTS:
                if (not self.item_count) or (not self.item_index):
                    return 0.0
                return min(1, float(self.item_index + 1) / float(self.item_count))
        return 0.0

    @computed_field  # type: ignore
    @property
    def artifact_progress_normalized(self) -> float:
        """Compute normalized artifact progress in range 0..1.

        Returns:
            float: The normalized artifact progress in range 0..1.
        """
        if self.status == DownloadProgressState.DOWNLOADING:
            if not self.artifact_size:
                return 0.0
            return min(1, float(self.artifact_downloaded_size) / float(self.artifact_size))
        if (
            has_qupath_extra
            and self.status == DownloadProgressState.QUPATH_ANNOTATE_INPUT_WITH_RESULTS
            and self.qupath_annotate_input_with_results_progress
        ):
            return self.qupath_annotate_input_with_results_progress.progress_normalized
        return 0.0


class Service(BaseService):
    """Service of the application module."""

    _settings: Settings
    _client: Client | None = None
    _platform_service: PlatformService | None = None

    def __init__(self) -> None:
        """Initialize service."""
        super().__init__(Settings)  # automatically loads and validates the settings

    def info(self, mask_secrets: bool = True) -> dict[str, Any]:
        """Determine info of this service.

        Args:
            mask_secrets (bool): If True, mask sensitive information in the output.

        Returns:
            dict[str,Any]: The info of this service.
        """
        return {"settings": self._settings.model_dump(context={UNHIDE_SENSITIVE_INFO: not mask_secrets})}

    def health(self) -> Health:  # noqa: PLR6301
        """Determine health of this service.

        Returns:
            Health: The health of the service.
        """
        return Health(
            status=Health.Code.UP,
        )

    def _get_platform_client(self) -> Client:
        """Get the platform client.

        Returns:
            Client: The platform client.

        Raises:
            Exception: If the client cannot be created.
        """
        if self._client is None:
            logger.debug("Creating platform client.")
            self._client = Client()
        else:
            logger.debug("Reusing platform client.")
        return self._client

    def _get_platform_service(self) -> PlatformService:
        """Get the platform service.

        Returns:
            PlatformService: The platform service.

        Raises:
            Exception: If the client cannot be created.
        """
        if self._platform_service is None:
            logger.debug("Creating platform service.")
            self._platform_service = PlatformService()
        else:
            logger.debug("Reusing platform service.")
        return self._platform_service

    def applications(self) -> list[Application]:
        """Get a list of all applications.

        Returns:
            list[str]: A list of all applications.

        Raises:
            Exception: If the client cannot be created.

        Raises:
            Exception: If the application list cannot be retrieved.
        """
        return [
            app
            for app in list(self._get_platform_client().applications.list())
            if app.application_id not in {"h-e-tme", "two-task-dummy"}
        ]

    def application(self, application_id: str) -> Application:
        """Get a specific application.

        Args:
            application_id (str): The ID of the application.

        Returns:
            Application: The application or None if not found.

        Raises:
            NotFoundException: If the application with the given ID is not found.
            Exception: If the application cannot be retrieved.
        """
        return self._get_platform_client().application(application_id)

    def application_version(
        self, application_version_id: str, use_latest_if_no_version_given: bool = False
    ) -> ApplicationVersion:
        """Get a specific application version.

        Args:
            application_version_id (str): The ID of the application version
            use_latest_if_no_version_given (bool): If True, use the latest version if no specific version is given.

        Returns:
            ApplicationVersion: The application version

        Raises:
            ValueError: If the application version ID is invalid.
            NotFoundException: If the application with the given ID is not found.
            Exception: If the application cannot be retrieved.
        """
        # Validate format: application_id:vX.Y.Z (where X.Y.Z is a semver)
        # This checks for proper format like "he-tme:v0.50.0" where "he-tme" is the application id
        # and "v0.50.0" is the version with proper semver format
        if not re.match(r"^[^:]+:v\d+\.\d+\.\d+$", application_version_id):
            if use_latest_if_no_version_given:
                latest_version = self.application_version_latest(self.application(application_version_id))
                if latest_version:
                    return latest_version
                message = f"No valid application version found for '{application_version_id}' "
                message += "and no latest version available."
                logger.warning(message)
                raise ValueError(message)
            message = f"Invalid application version id format: {application_version_id}. "
            "Expected format: application_id:vX.Y.Z"
            raise ValueError(message)

        application_id = application_version_id.split(":", maxsplit=1)[0]

        application = self.application(application_id)
        for version in self.application_versions(application):
            if version.application_version_id == application_version_id:
                return version
        message = f"Application version with ID {application_version_id} not found in application {application_id}"
        raise NotFoundException(message)

    def application_versions(self, application: Application) -> list[ApplicationVersion]:
        """Get a list of all versions of the given application.

        Args:
            application (Application): The application to check for versions.

        Returns:
            list[ApplicationVersion]: A list of all application versions.

        Raises:
            Exception: If version list cannot be retrieved
        """
        return self._get_platform_client().applications.versions.list_sorted(application=application)

    def application_version_latest(self, application: Application) -> ApplicationVersion | None:
        """Get a latest application version.

        Args:
            application (Application): The application to check for versions.

        Returns:
            ApplicationVersion | None: A list of all application versions.

        Raises:
            Exception: If version list cannot be retrieved
        """
        versions = self.application_versions(application)
        return versions[0] if versions else None

    @staticmethod
    def _process_key_value_pair(entry: dict[str, Any], key_value: str, reference: str) -> None:
        """Process a single key-value pair from a mapping.

        Args:
            entry: The entry dictionary to update
            key_value: String in the format "key=value"
            reference: The reference value for logging
        """
        key, value = key_value.split("=", 1)
        key = key.strip()
        value = value.strip()
        if not key:
            return

        if key not in entry:
            logger.warning("key '%s' not found in entry, ignoring mapping for '%s'", key, reference)
            return

        logger.debug("Updating key '%s' with value '%s' for reference '%s'.", key, value, reference)
        entry[key.strip()] = value.strip()

    @staticmethod
    def _apply_mappings_to_entry(entry: dict[str, Any], mappings: list[str]) -> None:
        """Apply key/value mappings to an entry.

        If the reference attribute of the entry matches the regex pattern in the mapping,
            the key/value pairs are applied.

        Args:
            entry: The entry dictionary to update with mapped values
            mappings: List of strings with format 'regex:key=value,...'
                where regex ismatched against the reference attribute in the entry
        """
        reference = entry["reference"]
        for mapping in mappings:
            parts = mapping.split(":", 1)
            if len(parts) != 2:  # noqa: PLR2004
                continue

            pattern = parts[0].strip()
            if not re.search(pattern, reference):
                continue

            key_value_pairs = parts[1].split(",")
            for key_value in key_value_pairs:
                Service._process_key_value_pair(entry, key_value, reference)

    @staticmethod
    def generate_metadata_from_source_directory(
        application_version_id: str,
        source_directory: Path,
        with_gui_metadata: bool = False,
        mappings: list[str] | None = None,
    ) -> list[dict[str, Any]]:
        """Generate metadata from the source directory.

        Steps:
        1. Recursively files ending with supported extensions in the source directory
        2. Creates a dict with the following columns
            - reference (str): The reference of the file, by default equivalent to the absolute file name
            - source (str): The absolute filename
            - checksum_base64_crc32c (str): The CRC32C checksum of the file constructed, base64 encoded
            - resolution_mpp (float): The microns per pixel, inspecting the base layer
            - height_px: The height of the image in pixels, inspecting the base layer
            - width_px: The width of the image in pixels, inspecting the base layer
            - Further attributes depending on the application and it's version
        3. Applies the optional mappings to fill in additional metadata fields in the dict.

        Args:
            application_version_id (str): The ID of the application version.
                If application id is given, the latest version of that application is used.
            source_directory (Path): The source directory to generate metadata from.
            with_gui_metadata (bool): If True, include additional metadata for GUI.
            mappings (list[str]): Mappings of the form '<regexp>:<key>:<value>,<key>:<value>,...'.
                The regular expression is matched against the reference attribute of the entry.
                The key/value pairs are applied to the entry if the pattern matches.

        Returns:
            dict[str, Any]: The generated metadata.

        Raises:
            Exception: If the metadata cannot be generated.

        Raises:
            ValueError: If the source directory does not exist or is not a directory.
        """
        logger.debug("Generating metadata from source directory: %s", source_directory)

        # TODO(Helmut): Use it
        application_version = Service().application_version(application_version_id, use_latest_if_no_version_given=True)  # noqa: F841

        metadata = []

        try:
            for extension in list(WSI_SUPPORTED_FILE_EXTENSIONS):
                for file_path in source_directory.glob(f"**/*{extension}"):
                    # Generate CRC32C checksum with google_crc32c and encode as base64
                    hash_sum = google_crc32c.Checksum()  # type: ignore[no-untyped-call]
                    with file_path.open("rb") as f:
                        while chunk := f.read(1024):
                            hash_sum.update(chunk)  # type: ignore[no-untyped-call]
                    checksum = str(base64.b64encode(hash_sum.digest()), "UTF-8")  # type: ignore[no-untyped-call]
                    try:
                        image_metadata = WSIService().get_metadata(file_path)
                        width = image_metadata["dimensions"]["width"]
                        height = image_metadata["dimensions"]["height"]
                        mpp = image_metadata["resolution"]["mpp_x"]
                        file_size_human = image_metadata["file"]["size_human"]
                        reference = file_path.absolute()
                        entry = {
                            "reference": str(reference),
                            "reference_short": str(reference.name),
                            "source": str(file_path),
                            "checksum_base64_crc32c": checksum,
                            "resolution_mpp": mpp,
                            "width_px": width,
                            "height_px": height,
                            "staining_method": None,
                            "tissue": None,
                            "disease": None,
                            "file_size_human": file_size_human,
                            "file_upload_progress": 0.0,
                            "platform_bucket_url": None,
                        }
                        if not with_gui_metadata:
                            entry.pop("reference_short", None)
                            entry.pop("source", None)
                            entry.pop("file_size_human", None)
                            entry.pop("file_upload_progress", None)

                        if mappings:
                            Service._apply_mappings_to_entry(entry, mappings)

                        metadata.append(entry)
                    except Exception:
                        message = f"Failed to process file '{file_path}'"
                        logger.exception(message)
                        continue

            logger.debug("Generated metadata for %d files", len(metadata))
            return metadata

        except Exception:
            logger.exception("Failed to generate metadata from source directory: %s", source_directory)
            raise

    @staticmethod
    def application_run_upload(
        application_version_id: str,
        metadata: list[dict[str, Any]],
        upload_prefix: str,
        upload_progress_queue: Any | None = None,  # noqa: ANN401
        upload_progress_callable: Callable[[int, Path, str], None] | None = None,
    ) -> bool:
        """Upload files with a progress queue.

        Args:
            application_version_id (str): The ID of the application version.
                If application id is given, the latest version of that application is used.
            metadata (list[dict[str, Any]]): The metadata to upload.
            upload_prefix (str): The prefix for the upload.
            upload_progress_queue (Queue | None): The queue to send progress updates to.
            upload_progress_callable (Callable[[int, Path, str], None] | None): The task to update for progress updates.

        Returns:
            bool: True if the upload was successful, False otherwise.
        """
        import psutil  # noqa: PLC0415

        logger.debug("Uploading files with upload ID '%s'", upload_prefix)
        application_version = Service().application_version(application_version_id, use_latest_if_no_version_given=True)
        for row in metadata:
            reference = row["reference"]
            source_file_path = Path(row["reference"])
            if not source_file_path.is_file():
                logger.warning("Source file '%s' does not exist.", row["referebce"])
                return False
            object_key = (
                f"{psutil.Process().username()}/{upload_prefix}/"
                f"{application_version.application_version_id}/{source_file_path.name}"
            )
            platform_bucket_url = (
                f"{BucketService().get_bucket_protocol()}://{BucketService().get_bucket_name()}/{object_key}"
            )
            signed_upload_url = BucketService().create_signed_upload_url(object_key)
            logger.debug("Generated signed upload URL '%s' for object '%s'", signed_upload_url, platform_bucket_url)
            if upload_progress_queue:
                upload_progress_queue.put_nowait({
                    "reference": reference,
                    "platform_bucket_url": platform_bucket_url,
                })
            file_size = source_file_path.stat().st_size
            logger.debug(
                "Uploading file '%s' with size %d bytes to '%s' via '%s'",
                source_file_path,
                file_size,
                platform_bucket_url,
                signed_upload_url,
            )
            with (
                open(source_file_path, "rb") as f,
            ):

                def read_in_chunks(  # noqa: PLR0913, PLR0917
                    reference: str,
                    file_size: int,
                    upload_progress_queue: Any | None = None,  # noqa: ANN401
                    upload_progress_callable: Callable[[int, Path, str], None] | None = None,
                    file_path: Path = source_file_path,
                    platform_bucket_url: str = platform_bucket_url,
                ) -> Generator[bytes, None, None]:
                    while True:
                        chunk = f.read(APPLICATION_RUN_UPLOAD_CHUNK_SIZE)
                        if not chunk:
                            break
                        if upload_progress_queue:
                            upload_progress_queue.put_nowait({
                                "reference": reference,
                                "file_upload_progress": min(100.0, f.tell() / file_size),
                            })
                        if upload_progress_callable:
                            upload_progress_callable(len(chunk), file_path, platform_bucket_url)
                        yield chunk

                response = requests.put(
                    signed_upload_url,
                    data=read_in_chunks(reference, file_size, upload_progress_queue, upload_progress_callable),
                    headers={"Content-Type": "application/octet-stream"},
                    timeout=60,
                )
                response.raise_for_status()
        logger.info("Upload completed successfully.")
        return True

    @staticmethod
    def application_runs_static(limit: int | None = None, completed_only: bool = False) -> list[dict[str, Any]]:
        return [
            {
                "application_run_id": run.application_run_id,
                "application_version_id": run.application_version_id,
                "triggered_at": run.triggered_at,
                "status": run.status,
            }
            for run in Service().application_runs(
                limit=limit, status=ApplicationRunStatus.COMPLETED if completed_only else None
            )
        ]

    def application_runs(
        self, limit: int | None = None, status: ApplicationRunStatus | None = None
    ) -> list[ApplicationRunData]:
        """Get a list of all application runs.

        Args:
            limit (int | None): The maximum number of runs to retrieve. If None, all runs are retrieved.
            status (ApplicationRunStatus | None): Filter runs by status. If None, all runs are retrieved.

        Returns:
            list[ApplicationRunData]: A list of all application runs.

        Raises:
            Exception: If the application run list cannot be retrieved.
        """
        if limit is not None and limit <= 0:
            return []
        runs = []
        page_size = LIST_APPLICATION_RUNS_MAX_PAGE_SIZE
        run_iterator = self._get_platform_client().runs.list_data(sort="-triggered_at", page_size=page_size)
        for run in run_iterator:
            if status is not None and run.status != status:
                continue
            runs.append(run)
            if limit is not None and len(runs) >= limit:
                break
        return runs

    def application_run(self, run_id: str) -> ApplicationRun:
        """Find a run by its ID.

        Args:
            run_id: The ID of the run to find

        Returns:
            ApplicationRun: The run that can be fetched using the .details() call.

        Raises:
            Exception: If initializing the client fails.
        """
        return self._get_platform_client().run(run_id)

    def application_run_submit_from_metadata(
        self, application_version_id: str, metadata: list[dict[str, Any]]
    ) -> ApplicationRun:
        """Submit a run for the given application.

        Args:
            application_version_id: The ID of the application version to run.
                If application id is given, the latest version of that application is used.
            metadata: The metadata for the run.

        Returns:
            ApplicationRun: The submitted run.

        Raises:
            ValueError: If platform bucket URL is missing or has unsupported protocol.
            Exception: If submitting the run failed unexpectedly.
        """
        logger.debug("Submitting application run with metadata: %s", metadata)
        items = []
        for row in metadata:
            platform_bucket_url = row["platform_bucket_url"]
            if platform_bucket_url and platform_bucket_url.startswith("gs://"):
                url_parts = platform_bucket_url[5:].split("/", 1)
                bucket_name = url_parts[0]
                object_key = url_parts[1]
                download_url = BucketService().create_signed_download_url(object_key, bucket_name)
            else:
                message = f"Invalid platform bucket URL: '{platform_bucket_url}'."
                logger.warning(message)
                raise ValueError(message)

            items.append(
                InputItem(
                    reference=row["reference"],
                    input_artifacts=[
                        InputArtifact(
                            name="user_slide",
                            download_url=download_url,
                            metadata={
                                "checksum_base64_crc32c": row["checksum_base64_crc32c"],
                                "height_px": int(row["height_px"]),
                                "width_px": int(row["width_px"]),
                                "media_type": (
                                    "image/tiff"
                                    if row["reference"].lower().endswith((".tif", ".tiff"))
                                    else "application/dicom"
                                    if row["reference"].lower().endswith(".dcm")
                                    else "application/octet-stream"
                                ),
                                "resolution_mpp": float(row["resolution_mpp"]),
                                "specimen": {
                                    "disease": row["disease"],
                                    "tissue": row["tissue"],
                                },
                                "staining_method": row["staining_method"],
                            },
                        )
                    ],
                )
            )
        logger.debug("Items for application run submission: %s", items)
        try:
            application_version = self.application_version(application_version_id, use_latest_if_no_version_given=True)
            run = self.application_run_submit(application_version.application_version_id, items)
            logger.info(
                "Submitted application run with items: %s, application run id %s", items, run.application_run_id
            )
            return run
        except Exception:
            logger.exception("Failed to submit application run.")
            raise

    def application_run_submit(self, application_version_id: str, items: list[InputItem]) -> ApplicationRun:
        """Submit a run for the given application.

        Args:
            application_version_id: The ID of the application version to run.
            items: The input items for the run.

        Returns:
            ApplicationRun: The submitted run.

        Raises:
            Exception: If submitting the run failed unexpectedly.
        """
        return self._get_platform_client().runs.create(application_version=application_version_id, items=items)

    def application_run_cancel(self, run_id: str) -> None:
        """Cancel a run by its ID.

        Args:
            run_id: The ID of the run to cancel

        Raises:
            Exception: If the client cannot be created.

        Raises:
            Exception: If canceling the run failed unexpectedly.
        """
        self.application_run(run_id).cancel()

    @staticmethod
    def application_run_download_static(  # noqa: PLR0913, PLR0917
        run_id: str,
        destination_directory: Path,
        create_subdirectory_for_run: bool = True,
        create_subdirectory_per_item: bool = True,
        wait_for_completion: bool = True,
        qupath_project: bool = False,
        download_progress_queue: Any | None = None,  # noqa: ANN401
    ) -> Path:
        return Service().application_run_download(
            run_id,
            destination_directory,
            create_subdirectory_for_run,
            create_subdirectory_per_item,
            wait_for_completion,
            qupath_project,
            download_progress_queue,
        )

    def application_run_download(  # noqa: C901, PLR0912, PLR0913, PLR0914, PLR0915, PLR0917
        self,
        run_id: str,
        destination_directory: Path,
        create_subdirectory_for_run: bool = True,
        create_subdirectory_per_item: bool = True,
        wait_for_completion: bool = True,
        qupath_project: bool = False,
        download_progress_queue: Any | None = None,  # noqa: ANN401
        download_progress_callable: Callable | None = None,  # type: ignore[type-arg]
    ) -> Path:
        """Download application run results with progress tracking.

        Args:
            progress (DownloadProgress): Progress tracking object for GUI or CLI updates.
            run_id (str): The ID of the application run to download.
            destination_directory (Path): Directory to save downloaded files.
            create_subdirectory_for_run (bool): Whether to create a subdirectory for the run.
            create_subdirectory_per_item (bool): Whether to create a subdirectory for each item,
                if not set, all items will be downloaded to the same directory but prefixed
                with the item reference and underscore.
            wait_for_completion (bool): Whether to wait for run completion. Defaults to True.
            qupath_project (bool): If True, create QuPath project referencing input slides and results.
                This requires QuPath to be installed. The QuPath project will be created in a subfolder
                of the destination directory.
            download_progress_queue (Queue | None): Queue for GUI progress updates.
            download_progress_callable (Callable | None): Callback for CLI progress updates.

        Returns:
            Path: The directory containing downloaded results.

        Raises:
            ValueError: If the run ID is invalid or destination directory cannot be created.
            RuntimeError: If run details cannot be retrieved or download fails.
            Exception: If run cannot be retrieved, destination directory cannot be created, or download fails.
        """
        if qupath_project and not has_qupath_extra:
            message = "QuPath project creation requested, but 'qupath' extra is not installed."
            message += 'Start launchpad with `uvx --with "aignostics[qupath]" ....'
            logger.warning(message)
            raise ValueError(message)
        progress = DownloadProgress()
        Service._update_progress(progress, download_progress_callable, download_progress_queue)

        application_run = self.application_run(run_id)
        final_destination_directory = destination_directory
        try:
            details = application_run.details()
        except ApiException as e:
            if e.status == HTTPStatus.UNPROCESSABLE_ENTITY:  # Don't use UNPROCESSABLE_CONTENT
                message = f"Run ID '{run_id}' invalid: {e!s}."
                logger.warning(message)
                raise ValueError(message) from e
            message = f"Failed to retrieve details for application run '{run_id}': {e}"
            logger.exception(message)
            raise RuntimeError(message) from e

        if create_subdirectory_for_run:
            final_destination_directory = destination_directory / details.application_run_id
        try:
            final_destination_directory.mkdir(parents=True, exist_ok=True)
        except OSError as e:
            message = f"Failed to create destination directory '{final_destination_directory}': {e}"
            logger.warning(message)
            raise ValueError(message) from e

        if qupath_project:

            def update_qupath_add_input_progress(qupath_add_input_progress: QuPathAddProgress) -> None:
                progress.status = DownloadProgressState.QUPATH_ADD_INPUT
                progress.qupath_add_input_progress = qupath_add_input_progress
                Service._update_progress(progress, download_progress_callable, download_progress_queue)

            logger.debug("Adding input slides to QuPath project ...")
            image_paths = []
            for item in application_run.results():
                image_path = Path(item.reference)
                if image_path.is_file():
                    image_paths.append(image_path.resolve())
            added = QuPathService.add(
                final_destination_directory / "qupath", image_paths, update_qupath_add_input_progress
            )
            message = f"Added '{added}' input slides to QuPath project."
            logger.info(message)

        logger.debug("Downloading results for run '%s' to '%s'", run_id, final_destination_directory)

        progress.status = DownloadProgressState.CHECKING
        Service._update_progress(progress, download_progress_callable, download_progress_queue)

        downloaded_items: set[str] = set()  # Track downloaded items to avoid re-downloading
        while True:
            run_details = application_run.details()  # (Re)load current run details
            progress.run = run_details
            Service._update_progress(progress, download_progress_callable, download_progress_queue)

            self._download_available_items(
                progress,
                application_run,
                final_destination_directory,
                downloaded_items,
                create_subdirectory_per_item,
                download_progress_queue,
                download_progress_callable,
            )

            if run_details.status in {
                ApplicationRunStatus.CANCELED_SYSTEM,
                ApplicationRunStatus.CANCELED_USER,
                ApplicationRunStatus.COMPLETED,
                ApplicationRunStatus.COMPLETED_WITH_ERROR,
                ApplicationRunStatus.COMPLETED_WITH_ERROR,
                ApplicationRunStatus.REJECTED,
            }:
                logger.debug("Run '%s' reached final status '%s'.", run_id, run_details.status)
                break

            if not wait_for_completion:
                logger.debug(
                    "Run '%s' is in progress with status '%s', but not requested to wait for completion.",
                    run_id,
                    run_details.status,
                )
                break

            logger.debug(
                "Run '%s' is in progress with status '%s', waiting for completion ...", run_id, run_details.status
            )
            progress.status = DownloadProgressState.WAITING
            Service._update_progress(progress, download_progress_callable, download_progress_queue)
            time.sleep(APPLICATION_RUN_DOWNLOAD_SLEEP_SECONDS)

        if qupath_project:
            logger.debug("Adding result images to QuPath project ...")

            def update_qupath_add_results_progress(qupath_add_results_progress: QuPathAddProgress) -> None:
                progress.status = DownloadProgressState.QUPATH_ADD_RESULTS
                progress.qupath_add_results_progress = qupath_add_results_progress
                Service._update_progress(progress, download_progress_callable, download_progress_queue)

            added = QuPathService.add(
                final_destination_directory / "qupath",
                [final_destination_directory],
                update_qupath_add_results_progress,
            )
            message = f"Added {added} result images to QuPath project."
            logger.info(message)
            logger.debug("Annotating input slides with polygons from results ...")

            def update_qupath_annotate_input_with_results_progress(
                qupath_annotate_input_with_results_progress: QuPathAnnotateProgress,
            ) -> None:
                progress.status = DownloadProgressState.QUPATH_ANNOTATE_INPUT_WITH_RESULTS
                progress.qupath_annotate_input_with_results_progress = qupath_annotate_input_with_results_progress
                Service._update_progress(progress, download_progress_callable, download_progress_queue)

            total_annotations = 0
            results = list(application_run.results())
            progress.item_count = len(results)
            for item_index, item in enumerate(application_run.results()):
                progress.item_index = item_index
                Service._update_progress(progress, download_progress_callable, download_progress_queue)

                image_path = Path(item.reference)
                if not image_path.is_file():
                    continue
                for artifact in item.output_artifacts:
                    if (
                        get_mime_type_for_artifact(artifact) == "application/geo+json"
                        and artifact.name == "cell_classification:geojson_polygons"
                    ):
                        artifact_name = artifact.name
                        if create_subdirectory_per_item:
                            reference_path = Path(item.reference)
                            stem_name = reference_path.stem
                            artifact_path = (
                                final_destination_directory
                                / stem_name
                                / f"{sanitize_path_component(artifact_name)}.json"
                            )
                        else:
                            artifact_path = (
                                final_destination_directory / f"{sanitize_path_component(artifact_name)}.json"
                            )
                        message = f"Annotating input slide '{image_path}' with artifact '{artifact_path}' ..."
                        logger.debug(message)
                        added = QuPathService.annotate(
                            final_destination_directory / "qupath",
                            image_path,
                            artifact_path,
                            update_qupath_annotate_input_with_results_progress,
                        )
                        message = f"Added {added} annotations to input slide '{image_path}' from '{artifact_path}'."
                        logger.info(message)
                        total_annotations += added
            message = f"Added {added} annotations to input slides."
            logger.info(message)

        progress.status = DownloadProgressState.COMPLETED
        Service._update_progress(progress, download_progress_callable, download_progress_queue)

        return final_destination_directory

    @staticmethod
    def _update_progress(
        progress: DownloadProgress,
        download_progress_callable: Callable | None = None,  # type: ignore[type-arg]
        download_progress_queue: Any | None = None,  # noqa: ANN401
    ) -> None:
        if download_progress_callable:
            download_progress_callable(progress)
        if download_progress_queue:
            download_progress_queue.put_nowait(progress)

    def _download_available_items(  # noqa: PLR0913, PLR0917
        self,
        progress: DownloadProgress,
        application_run: ApplicationRun,
        destination_directory: Path,
        downloaded_items: set[str],
        create_subdirectory_per_item: bool = False,
        download_progress_queue: Any | None = None,  # noqa: ANN401
        download_progress_callable: Callable | None = None,  # type: ignore[type-arg]
    ) -> None:
        """Download items that are available and not yet downloaded.

        Args:
            progress (DownloadProgress): Progress tracking object for GUI or CLI updates.
            application_run (ApplicationRun): The application run object.
            destination_directory (Path): Directory to save files.
            downloaded_items (set): Set of already downloaded item references.
            create_subdirectory_per_item (bool): Whether to create a subdirectory for each item.
            download_progress_queue (Queue | None): Queue for GUI progress updates.
            download_progress_callable (Callable | None): Callback for CLI progress updates.
        """
        items = list(application_run.results())
        progress.item_count = len(items)
        for item_index, item in enumerate(items):
            if item.reference in downloaded_items:
                continue

            if item.status == ItemStatus.SUCCEEDED:
                progress.status = DownloadProgressState.DOWNLOADING
                progress.item_index = item_index
                progress.item = item
                progress.item_reference = item.reference

                progress.artifact_count = len(item.output_artifacts)
                Service._update_progress(progress, download_progress_callable, download_progress_queue)

                if create_subdirectory_per_item:
                    reference_path = Path(item.reference)
                    stem_name = reference_path.stem
                    try:
                        # Handle case where reference might be relative to destination
                        rel_path = reference_path.relative_to(destination_directory)
                        stem_name = rel_path.stem
                    except ValueError:
                        # Not a subfolder - just use the stem
                        pass
                    item_directory = destination_directory / stem_name
                else:
                    item_directory = destination_directory
                item_directory.mkdir(exist_ok=True)

                for artifact_index, artifact in enumerate(item.output_artifacts):
                    progress.artifact_index = artifact_index
                    progress.artifact = artifact
                    Service._update_progress(progress, download_progress_callable, download_progress_queue)

                    self._download_item_artifact(
                        progress,
                        artifact,
                        item_directory,
                        item.reference if not create_subdirectory_per_item else "",
                        download_progress_queue,
                        download_progress_callable,
                    )

                downloaded_items.add(item.reference)

    def _download_item_artifact(  # noqa: PLR0913, PLR0917
        self,
        progress: DownloadProgress,
        artifact: Any,  # noqa: ANN401
        destination_directory: Path,
        prefix: str = "",
        download_progress_queue: Any | None = None,  # noqa: ANN401
        download_progress_callable: Callable | None = None,  # type: ignore[type-arg]
    ) -> None:
        """Download a an artifact of a result item with progress tracking.

        Args:
            progress (DownloadProgress): Progress tracking object for GUI or CLI updates.
            artifact (Any): The artifact to download.
            destination_directory (Path): Directory to save the file.
            prefix (str): Prefix for the file name, if needed.
            download_progress_queue (Queue | None): Queue for GUI progress updates.
            download_progress_callable (Callable | None): Callback for CLI progress updates.

        Raises:
            ValueError: If no checksum metadata is found for the artifact.
            requests.HTTPError: If the download fails.
        """
        metadata = artifact.metadata or {}
        metadata_checksum = metadata.get("checksum_base64_crc32c", "") or metadata.get("checksum_crc32c", "")
        if not metadata_checksum:
            message = f"No checksum metadata found for artifact {artifact.name}"
            logger.error(message)
            raise ValueError(message)

        artifact_path = (
            destination_directory
            / f"{prefix}{sanitize_path_component(artifact.name)}{get_file_extension_for_artifact(artifact)}"
        )

        if artifact_path.exists():
            checksum = google_crc32c.Checksum()  # type: ignore[no-untyped-call]
            with open(artifact_path, "rb") as f:
                while chunk := f.read(APPLICATION_RUN_FILE_READ_CHUNK_SIZE):
                    checksum.update(chunk)  # type: ignore[no-untyped-call]
            existing_checksum = base64.b64encode(checksum.digest()).decode("ascii")  # type: ignore[no-untyped-call]
            if existing_checksum == metadata_checksum:
                logger.debug("File %s already exists with correct checksum", artifact_path)
                return

        self._download_file_with_progress(
            progress,
            artifact.download_url,
            artifact_path,
            metadata_checksum,
            download_progress_queue,
            download_progress_callable,
        )

    @staticmethod
    def _download_file_with_progress(  # noqa: PLR0913, PLR0917
        progress: DownloadProgress,
        signed_url: str,
        artifact_path: Path,
        metadata_checksum: str,
        download_progress_queue: Any | None = None,  # noqa: ANN401
        download_progress_callable: Callable | None = None,  # type: ignore[type-arg]
    ) -> None:
        """Download a file with progress tracking support.

        Args:
            progress (DownloadProgress): Progress tracking object for GUI or CLI updates.
            signed_url (str): The signed URL to download from.
            artifact_path (Path): Path to save the file.
            metadata_checksum (str): Expected CRC32C checksum in base64.
            download_progress_queue (Any | None): Queue for GUI progress updates.
            download_progress_callable (Callable | None): Callback for CLI progress updates.

        Raises:
            ValueError: If checksum verification fails.
            requests.HTTPError: If download fails.
        """
        logger.debug(
            "Downloading artifact '%s' to '%s' with expected checksum '%s' for item reference '%s'",
            progress.artifact.name if progress.artifact else "unknown",
            artifact_path,
            metadata_checksum,
            progress.item_reference or "unknown",
        )
        progress.artifact_download_url = signed_url
        progress.artifact_path = artifact_path
        progress.artifact_downloaded_size = 0
        progress.artifact_downloaded_chunk_size = 0
        progress.artifact_size = None
        Service._update_progress(progress, download_progress_callable, download_progress_queue)

        checksum = google_crc32c.Checksum()  # type: ignore[no-untyped-call]

        with requests.get(signed_url, stream=True, timeout=60) as stream:
            stream.raise_for_status()
            progress.artifact_size = int(stream.headers.get("content-length", 0))
            Service._update_progress(progress, download_progress_callable, download_progress_queue)
            with open(artifact_path, mode="wb") as file:
                for chunk in stream.iter_content(chunk_size=APPLICATION_RUN_DOWNLOAD_CHUNK_SIZE):
                    if chunk:
                        file.write(chunk)
                        checksum.update(chunk)  # type: ignore[no-untyped-call]
                        progress.artifact_downloaded_chunk_size = len(chunk)
                        progress.artifact_downloaded_size += progress.artifact_downloaded_chunk_size
                        Service._update_progress(progress, download_progress_callable, download_progress_queue)

        downloaded_checksum = base64.b64encode(checksum.digest()).decode("ascii")  # type: ignore[no-untyped-call]
        if downloaded_checksum != metadata_checksum:
            artifact_path.unlink()  # Remove corrupted file
            msg = f"Checksum mismatch for {artifact_path}: {downloaded_checksum} != {metadata_checksum}"
            logger.error(msg)
            raise ValueError(msg)



"""Notebook module."""

from importlib.util import find_spec

__all__ = []

if find_spec("marimo") and find_spec("nicegui"):
    from ._gui import PageBuilder
    from ._service import Service

    __all__ += [
        "PageBuilder",
        "Service",
    ]



"""Service of the wsi module."""

import io
from pathlib import Path
from typing import Any

import requests

from aignostics import WSI_SUPPORTED_FILE_EXTENSIONS
from aignostics.utils import BaseService, Health, get_logger

logger = get_logger(__name__)

TIMEOUT = 60  # 1 minutes


class Service(BaseService):
    """Service of the application module."""

    def info(self, mask_secrets: bool = True) -> dict[str, Any]:  # noqa: ARG002, PLR6301
        """Determine info of this service.

        Args:
            mask_secrets (bool): Whether to mask sensitive information in the output.

        Returns:
            dict[str,Any]: The info of this service.
        """
        return {}

    def health(self) -> Health:  # noqa: PLR6301
        """Determine health of thumbnail service.

        Returns:
            Health: The health of the service.
        """
        return Health(
            status=Health.Code.UP,
        )

    @staticmethod
    def get_thumbnail(path: Path) -> "PIL.Image.Image":  # type: ignore # noqa: F821
        """Get thumbnail as PIL image.

        Args:
            path (Path): Path to the image.

        Returns:
            PIL.Image.Image: Thumbnail of the image.

        Raises:
            ValueError: If the file type is not supported.
            RuntimeError: If there is an error generating the thumbnail.
        """
        from ._openslide_handler import OpenSlideHandler  # noqa: PLC0415

        if path.exists() is False:
            message = f"File does not exist: {path}"
            logger.warning(message)
            raise ValueError(message)
        if path.suffix.lower() not in WSI_SUPPORTED_FILE_EXTENSIONS:
            message = f"Unsupported file type: {path.suffix}. Supported types are {WSI_SUPPORTED_FILE_EXTENSIONS!s}"
            logger.warning(message)
            raise ValueError(message)
        try:
            return OpenSlideHandler.from_file(path).get_thumbnail()
        except Exception as e:
            message = f"Error processing file {path}: {e!s}"
            logger.exception(message)
            raise RuntimeError(message) from e

    @staticmethod
    def get_thumbnail_bytes(path: Path) -> bytes:
        """Get thumbnail of a image as bytes.

        Args:
            path (Path): Path to the image.

        Returns:
            bytes: Thumbnail of the image.

        Raises:
            ValueError: If the file type is not supported.
            RuntimeError: If there is an error processing the file.
        """
        thumbnail_image = Service.get_thumbnail(path)
        buffer = io.BytesIO()
        thumbnail_image.save(buffer, format="PNG")
        return buffer.getvalue()

    @staticmethod
    def get_metadata(path: Path) -> dict[str, Any]:
        """Get metadata from a TIFF file.

        Args:
            path (Path): Path to the TIFF file.

        Returns:
            dict[str, Any]: Metadata of the TIFF file.

        Raises:
            ValueError: If the file type is not supported.
            RuntimeError: If there is an error processing the file.
        """
        from ._openslide_handler import OpenSlideHandler  # noqa: PLC0415

        if path.exists() is False:
            message = f"File does not exist: {path}"
            logger.warning(message)
            raise ValueError(message)
        if path.suffix.lower() not in WSI_SUPPORTED_FILE_EXTENSIONS:
            message = f"Unsupported file type: {path.suffix}. Supported types are {WSI_SUPPORTED_FILE_EXTENSIONS}."
            logger.warning(message)
            raise ValueError(message)
        try:
            return OpenSlideHandler.from_file(path).get_metadata()
        except Exception as e:
            error_msg = f"Error processing file {path}: {e!s}"
            logger.exception(error_msg)
            raise RuntimeError(error_msg) from e

    @staticmethod
    def get_tiff_as_jpg(url: str) -> bytes:
        """Get a TIFF image from a URL and convert it to JPG format.

        Args:
            url (str): URL to the TIFF image.

        Returns:
            bytes: The TIFF image converted to JPG format as bytes.

        Raises:
            ValueError: If URL format is invalid or if there's an error opening the tiff.
            RuntimeError: If there's an unexpected internal error.
        """
        from PIL import Image as PILImage  # noqa: PLC0415
        from PIL import UnidentifiedImageError  # noqa: PLC0415

        if not url.startswith(("http://localhost", "https://")):
            error_msg = "URL must start with 'http://localhost' or 'https://'."
            logger.warning(error_msg)
            raise ValueError(error_msg)
        try:
            response = requests.get(url, timeout=TIMEOUT)
            response.raise_for_status()
            tiff_data = response.content
            tiff_buffer = io.BytesIO(tiff_data)
            with PILImage.open(tiff_buffer) as img:
                rgb_img = img.convert("RGB") if img.mode != "RGB" else img
                jpg_buffer = io.BytesIO()
                rgb_img.save(jpg_buffer, format="JPEG", quality=90)
                return jpg_buffer.getvalue()
        except requests.HTTPError as e:
            error_msg = f"HTTP error while fetching TIFF from URL: {e!s}."
            logger.warning(error_msg)
            raise ValueError(error_msg) from e
        except requests.exceptions.InvalidURL as e:
            error_msg = f"URL error prevented fetching TIFF: {e!s}."
            logger.warning(error_msg)
            raise ValueError(error_msg) from e
        except requests.URLRequired as e:
            error_msg = f"URL error prevented fetching TIFF: {e!s}."
            logger.warning(error_msg)
            raise ValueError(error_msg) from e
        except UnidentifiedImageError as e:
            error_msg = f"Unidentified image error while trying to process as TIFF: {e!s}."
            logger.warning(error_msg)
            raise ValueError(error_msg) from e
        except Exception as e:
            error_msg = f"Unexpected error converting TIFF to JPEG: {e!s}."
            logger.exception(error_msg)
            raise RuntimeError(error_msg) from e



from .item_result_read_response import *
from .validation_error_loc_inner import *
from .application_version_read_response import *
from .item_status import *
from .run_creation_response import *
from .input_artifact_read_response import *
from .user_payload import *
from .validation_error import *
from .application_read_response import *
from .output_artifact_scope import *
from .input_artifact_creation_request import *
from .item_creation_request import *
from .http_validation_error import *
from .transfer_urls import *
from .output_artifact_result_read_response import *
from .run_read_response import *
from .application_run_status import *
from .run_creation_request import *
from .payload_output_artifact import *
from .payload_input_artifact import *
from .payload_item import *
from .output_artifact_read_response import *




"""
    Aignostics Platform API reference

     Pagination is done via `page` and `page_size`. Sorting via `sort` query parameter. The `sort` query parameter can be provided multiple times. The sorting direction can be indicated via `+` (ascending) or `-` (descending) (e.g. `/v1/applications?sort=+name)`.

    The version of the OpenAPI document: 1.0.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from datetime import datetime
from pydantic import BaseModel, ConfigDict, Field, StrictStr
from typing import Any, ClassVar, Dict, List, Optional
from aignx.codegen.models.application_run_status import ApplicationRunStatus
from aignx.codegen.models.user_payload import UserPayload
from typing import Optional, Set
from typing_extensions import Self

class RunReadResponse(BaseModel):
    """
    RunReadResponse
    """ # noqa: E501
    application_run_id: StrictStr = Field(description="UUID of the application")
    application_version_id: StrictStr = Field(description="ID of the application version")
    organization_id: StrictStr = Field(description="Organization of the owner of the application run")
    user_payload: Optional[UserPayload] = None
    status: ApplicationRunStatus = Field(description=" When the application run request is received by the Platform, the `status` of it is set to `received`. Then it is transitioned to `scheduled`, when it is scheduled for the processing. When the application run is scheduled, it will process the input items and generate the result incrementally. As soon as the first result is generated, the state is changed to `running`. The results can be downloaded via `/v1/runs/{run_id}/results` endpoint. When all items are processed and all results are generated, the application status is set to `completed`. If the processing is done, but some items fail, the status is set to `completed_with_error`.  When the application run request is rejected by the Platform before scheduling, it is transferred to `rejected`. When the application run reaches the threshold of number of failed items, the whole application run is set to `canceled_system` and the remaining pending items are not processed. When the application run fails, the finished item results are available for download.  If the application run is canceled by calling `POST /v1/runs/{run_id}/cancel` endpoint, the processing of the items is stopped, and the application status is set to `cancelled_user`     ")
    triggered_at: datetime = Field(description="Timestamp showing when the application run was triggered")
    triggered_by: StrictStr = Field(description="Id of the user who triggered the application run")
    __properties: ClassVar[List[str]] = ["application_run_id", "application_version_id", "organization_id", "user_payload", "status", "triggered_at", "triggered_by"]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of RunReadResponse from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([
        ])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of user_payload
        if self.user_payload:
            _dict['user_payload'] = self.user_payload.to_dict()
        # set to None if user_payload (nullable) is None
        # and model_fields_set contains the field
        if self.user_payload is None and "user_payload" in self.model_fields_set:
            _dict['user_payload'] = None

        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of RunReadResponse from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "application_run_id": obj.get("application_run_id"),
            "application_version_id": obj.get("application_version_id"),
            "organization_id": obj.get("organization_id"),
            "user_payload": UserPayload.from_dict(obj["user_payload"]) if obj.get("user_payload") is not None else None,
            "status": obj.get("status"),
            "triggered_at": obj.get("triggered_at"),
            "triggered_by": obj.get("triggered_by")
        })
        return _obj



"""GUI of application module including homepage of Aignostics Launchpad."""

from pathlib import Path

from aignostics.utils import BasePageBuilder


class PageBuilder(BasePageBuilder):
    @staticmethod
    def register_pages() -> None:
        from nicegui import Client, app, ui  # noq  # noqa: PLC0415

        app.add_static_files("/application_assets", Path(__file__).parent / "assets")

        @ui.page("/")
        async def page_index(client: Client) -> None:
            """Index page of application module, serving as the homepage of Aignostics Launchpad."""
            from ._page_index import _page_index  # noqa: PLC0415

            await _page_index(client)

        @ui.page("/application/{application_id}")
        async def page_application_describe(application_id: str) -> None:
            """Describe Application.

            Args:
                application_id (str): The application ID.
            """
            from ._page_application_describe import _page_application_describe  # noqa: PLC0415

            await _page_application_describe(application_id)

        @ui.page("/application/run/{application_run_id}")
        async def page_application_run_describe(application_run_id: str) -> None:
            """Describe Application Run.

            Args:
                application_run_id (str): The application run id
            """
            from ._page_application_run_describe import _page_application_run_describe  # noqa: PLC0415

            await _page_application_run_describe(application_run_id)



# /// script
# requires-python = ">=3.13"
# dependencies = [
#     "marimo",
#     "aignostics",
# ]
# ///

import marimo

__generated_with = "0.13.15"
app = marimo.App(width="full")


@app.cell
def _():
    """Aignostics Notebook

    Notes:
    - See https://colab.research.google.com/github/ImagingDataCommons/IDC-Tutorials/blob/master/notebooks/pathomics/microscopy_dicom_ann_intro.ipynb#scrollTo=fMRsnFlzinO5

    """
    import os
    from pathlib import Path

    import marimo as mo
    import matplotlib.pyplot as plt
    from cloudpathlib import GSPath
    from dotenv import load_dotenv
    from aignostics import dataset, WSI_SUPPORTED_FILE_EXTENSIONS
    from wsidicom import WsiDicom
    from aignostics.utils import get_user_data_directory

    mo.sidebar([
        mo.md("# aignostics"),
        mo.nav_menu(
            {
                "#/home": f"{mo.icon('lucide:home')} Home",
                "#/about": f"{mo.icon('lucide:user')} Help",
                "Links": {
                    "https://platform.aignostics.com": "Platform",
                    "https://github.com/aignotics/python-sdk": "GitHub",
                },
            },
            orientation="vertical",
        ),
    ])


@app.cell
def _(Path, mo):
    query_params = mo.query_params()
    application_run_id = query_params.get("application_run_id","Not set")
    print(f"Application run id '{application_run_id}'")
    mo.vstack([
        mo.ui.file_browser(
            Path(query_params.get("results_folder", get_user_data_directory("results"))),
            filetypes=list(WSI_SUPPORTED_FILE_EXTENSIONS),
            multiple=True,
            restrict_navigation=True,
        )
    ])
    return


@app.cell
def _(dataset):
    idc_client = dataset.IDCClient()  # set-up idc_client
    idc_client.fetch_index("sm_instance_index")
    return (idc_client,)


@app.cell
def _(idc_client):
    query_sr = """
    SELECT
        SeriesInstanceUID,
        StudyInstanceUID,
        PatientID,
        collection_id
    FROM
        index
    WHERE
        analysis_result_id = 'Pan-Cancer-Nuclei-Seg-DICOM' AND Modality = 'ANN' AND collection_id = 'tcga_luad'
    ORDER BY
        crdc_series_uuid
    LIMIT 1
    """
    pan_ann = idc_client.sql_query(query_sr)
    return (pan_ann,)


@app.cell
def _(idc_client, pan_ann):
    study_instance_id = pan_ann["StudyInstanceUID"].iloc[0]
    viewer_url = idc_client.get_viewer_URL(
        studyInstanceUID=study_instance_id, viewer_selector="slim"
    )
    from IPython.display import IFrame

    IFrame(viewer_url, width=1260, height=900)
    return


@app.cell
def _(Path, WsiDicom, idc_client, plt):
    series_instance_uid = "1.3.6.1.4.1.5962.99.1.1069745200.1645485340.1637452317744.2.0"  # copied from slimviewer info box
    idc_client.download_from_selection(
        seriesInstanceUID=series_instance_uid,
        downloadDir="tmp/",
        dirTemplate="%SeriesInstanceUID",
        use_s5cmd_sync=True,
    )
    print(Path(f"tmp/{series_instance_uid}").exists())
    slide = WsiDicom.open(f"tmp/{series_instance_uid}")

    fig, axes = plt.subplots(1, 1, figsize=(12, 6))
    thumbnail = slide.read_thumbnail()
    print(axes)
    axes.imshow(thumbnail)
    axes.axis("off")
    plt.show()
    return


@app.cell
def _():
    return


if __name__ == "__main__":
    app.run()
